# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Flax functional core: Scopes."""

import collections
import contextlib
import dataclasses
import functools
import hashlib
import typing
from typing import (
  Any,
  Callable,
  Dict,
  Generic,
  Iterable,
  Literal,
  Mapping,
  Optional,
  Sequence,
  Set,
  Tuple,
  TypeVar,
  Union,
  cast,
  overload,
)

import jax
import numpy as np
from jax import numpy as jnp
from jax import random, tree_util

from flax import config as config
from flax import configurations as legacy_config  # only for flax_lazy_rng
from flax import errors, struct, traceback_util
from flax.ids import uuid
from flax.typing import (
  PRNGKey,
  Array,
  RNGSequences,
  Collection,
  MutableCollection,
  VariableDict,
  FrozenVariableDict as FrozenVariableDict,
  MutableVariableDict,
  PRNGFoldable,
)

from . import meta, partial_eval, tracers
from .frozen_dict import FrozenDict, freeze, unfreeze

traceback_util.register_exclusion(__file__)

T = TypeVar('T')


Filter = Union[bool, str, typing.Collection[str], 'DenyList']

# When conditioning on filters we require explicit boolean comparisons.
# pylint: disable=g-bool-id-comparison


@dataclasses.dataclass(frozen=True, eq=True)
class DenyList:
  """DenyList represents an opt-out based mutability filter.
  DenyList can be used to make every collection mutable except the ones
  defined in the given filter.
  To for example make everything but the params collection mutable::
    nn.apply(fn, mutable=nn.DenyList(["params"]))
  Attributes:
    deny: The filter representing the collections that are not mutable.
  """

  deny: Filter


CollectionFilter = Filter
PRNGSequenceFilter = Filter


class LazyRng(struct.PyTreeNode):
  """Wrapper around JAX PRNGKey that lazily maintains a tuple of static data to be folded into the rng."""

  rng: PRNGKey
  suffix: Tuple[PRNGFoldable, ...] = struct.field(pytree_node=False)

  def as_jax_rng(self) -> PRNGKey:
    return _fold_in_static(self.rng, self.suffix)

  @staticmethod
  def create(
    rng: Union['LazyRng', PRNGKey], *suffix: PRNGFoldable
  ) -> 'LazyRng':
    if not legacy_config.flax_lazy_rng:
      if isinstance(rng, LazyRng):
        assert not rng.suffix
        rng = rng.rng
      return LazyRng(_legacy_rng_fold_in(rng, suffix), ())
    if isinstance(rng, LazyRng):
      return LazyRng(rng.rng, rng.suffix + suffix)
    else:
      return LazyRng(rng, suffix)


def _legacy_rng_fold_in(rng: PRNGKey, data: Iterable[PRNGFoldable]) -> PRNGKey:
  """Legacy RNG folding."""
  for x in data:
    if isinstance(x, str):
      m = hashlib.sha1()
      m.update(x.encode('utf-8'))
      d = m.digest()
      hash_int = int.from_bytes(d[:4], byteorder='big')
      rng = random.fold_in(rng, jnp.uint32(hash_int))  # type: ignore
    elif isinstance(x, int):
      rng = random.fold_in(rng, x)
    else:
      raise ValueError(f'Expected int or string, got: {x}')
  return rng


def _fold_in_static(
  rng: PRNGKey, data: typing.Collection[PRNGFoldable]
) -> PRNGKey:
  """Folds static data (strings & ints) into a jax.random.PRNGKey using its SHA-1 hash.

  This is faster than splitting an PRNGKey because it allows generating new PRNG
  keys in parallel that are independent of each other.

  Args:
   rng: the rng to fold the string into.
   data: the string to be folded in.

  Returns:
   The newly generated PRNG key.
  """
  if not data:
    return rng
  m = hashlib.sha1()
  for x in data:
    if config.flax_fix_rng_separator:
      # encode seperate to avoid collisions like for example: ("ab", "c") and ("a", "bc")
      m.update(b'\00')
    if isinstance(x, str):
      m.update(x.encode('utf-8'))
    elif isinstance(x, int):
      m.update(x.to_bytes((x.bit_length() + 7) // 8, byteorder='big'))
    else:
      raise ValueError(f'Expected int or string, got: {x}')
  d = m.digest()
  hash_int = int.from_bytes(d[:4], byteorder='big')
  return random.fold_in(rng, jnp.uint32(hash_int))  # type: ignore


def is_filter_empty(filter_like: Filter) -> bool:
  """Returns True if `filter_like` is an empty filter.

  Args:
    filter_like: The filter to test.

  Returns:
    A filter is empty when it is an empty collection, it is a bool with value
    False, ir it is a DenyList that matches everything. A string filter is never
    empty.
  """
  if isinstance(filter_like, str):
    return False
  if isinstance(filter_like, typing.Collection):
    return not filter_like
  if isinstance(filter_like, bool):
    return not filter_like
  if isinstance(filter_like, DenyList):
    # if any arbitrary collection is in the denylist it matches everything so
    # the filter is empty. This is checked with a stub.
    return in_filter(filter_like.deny, '__flax_internal_stub__')
  raise errors.InvalidFilterError(filter_like)


def in_filter(filter_like: Filter, col: str) -> bool:
  """Checks whether a filter can be applied to a collection.

  Used for both collections and rng sequence filters.

  Args:
    filter_like: a filter (either a boolean, a string, or a list of strings) for
      a collection.
    col: a collection, which is a string identifying a dictionary of data, for
      instance "params" or "batch_stats".

  Returns:
    True if either `filter_like` is True, equal to `col`, or a sequence
    containing `col`.
  """
  if isinstance(filter_like, str):
    return col == filter_like
  if isinstance(filter_like, typing.Collection):
    return col in filter_like
  if isinstance(filter_like, bool):
    return filter_like
  if isinstance(filter_like, DenyList):
    return not in_filter(filter_like.deny, col)
  raise errors.InvalidFilterError(filter_like)


def filter_to_set(x: Filter) -> Set[str]:
  """Converts a Filter into a set of collections, fails on the infinite set.

  Args:
    x: a filter (boolean, string, or list of strings).

  Returns:
    The input filter represented as a set of strings.
  """
  assert x is not True and not isinstance(x, DenyList), 'Infinite set'
  if x is False:
    return set()
  if isinstance(x, str):
    return set([x])
  if isinstance(x, typing.Collection):
    return set(x)
  raise errors.InvalidFilterError(x)


def union_filters(a: Filter, b: Filter) -> Filter:
  """Takes the union of two filters (similar to a logical or).

  Args:
    a: a filter.
    b: a filter.

  Returns:
    The union of the two input filters. For instance,
    `union_filters('f1', ['f2']) = {'f1', 'f2'}`.
  """
  if a is True or b is True:
    return True
  if isinstance(a, DenyList) and isinstance(b, DenyList):
    return DenyList(intersect_filters(a.deny, b.deny))
  if isinstance(b, DenyList):
    a, b = b, a
  if isinstance(a, DenyList):
    return DenyList(subtract_filters(a.deny, b))

  a = filter_to_set(a)
  b = filter_to_set(b)
  return a.union(b)


def subtract_filters(a: Filter, b: Filter) -> Filter:
  """Returns the subtraction of b from a.

  Args:
    a: a filter.
    b: a filter.

  Returns:
    A filter matching with values in a that are not in b.
  """
  if b is True:
    return False
  if a is True:
    return DenyList(b)
  if isinstance(a, DenyList) and isinstance(b, DenyList):
    return subtract_filters(b.deny, a.deny)
  if isinstance(a, DenyList):
    return DenyList(union_filters(a.deny, b))
  if isinstance(b, DenyList):
    return intersect_filters(a, b.deny)
  a = filter_to_set(a)
  b = filter_to_set(b)
  return a - b


def intersect_filters(a: Filter, b: Filter) -> Filter:
  """Take the intersection of two filters (similar to a logical and).

  Args:
    a: a filter.
    b: a filter.

  Returns:
    The intersection of the two input filters. For instance,
    `intersect_filters('f1', ['f1', 'f2']) = {'f1'}`.
  """
  if a is True:
    return b
  if b is True:
    return a
  if isinstance(a, DenyList) and isinstance(b, DenyList):
    return DenyList(union_filters(b.deny, a.deny))
  if isinstance(b, DenyList):
    b, a = a, b
  if isinstance(a, DenyList):
    return subtract_filters(b, a.deny)
  a = filter_to_set(a)
  b = filter_to_set(b)
  return a.intersection(b)


def group_collections(
  xs: VariableDict, col_filters: Sequence[CollectionFilter]
) -> Sequence[MutableVariableDict]:
  """Groups variables by collection filters.

  Iteratively applies the filters in `col_filters` to `xs`, and adds the result
  of applying each filter to the output sequence. Each key in `xs` is only added
  to the output once.

  Args:
    xs: a dictionary of variables, keyed by collections (strings).
    col_filters: a list of collection filters.

  Returns:
    A sequence S with `len(S) == len(col_filters)`. Each `S[i]` is the result of
    applying filter `col_filters[i]` to the remaining keys in `xs`.
  """
  cols: Iterable[str]
  cols = xs.keys()
  groups = []
  for col_filter in col_filters:
    remaining_cols = []
    group = {}
    for col in cols:
      if in_filter(col_filter, col):
        group[col] = jax.tree_util.tree_map(lambda x: x, xs[col])
      else:
        remaining_cols.append(col)
    cols = remaining_cols
    groups.append(group)
  return tuple(groups)


class Variable(Generic[T]):
  """A Variable object allows mutable access to a variable in a VariableDict.

  Variables are identified by a collection (e.g., "batch_stats") and a name
  (e.g., "moving_mean"). The value property gives access to the variable's
  content and can be assigned to for mutation.
  """

  def __init__(self, scope: 'Scope', collection: str, name: str, unbox: bool):
    """Initializes a variable.

    Args:
      scope: The scope in which the variable is stored.
      collection: The collection of the variable (e.g., "params").
      name: The name of the variable (e.g., "dense").
      unbox: Whether to unbox boxed values with metadata.
    """
    self._id = uuid()
    self.scope = scope
    self.collection = collection
    self.name = name
    self.unbox = unbox

  @property
  def value(self) -> T:
    """Returns the value of this Variable."""
    v = self.scope.get_variable(self.collection, self.name)
    return meta.unbox(v) if self.unbox else v

  @value.setter
  def value(self, value: T):
    """Updates the value of this Variable."""
    if self.unbox:
      cur = self.scope.get_variable(self.collection, self.name)
      cur_struct = tree_util.tree_structure(cur, is_leaf=meta.is_axis_metadata)
      value_struct = tree_util.tree_structure(
        value, is_leaf=meta.is_axis_metadata
      )
      has_meta = any(map(meta.is_axis_metadata, cur_struct.flatten_up_to(cur)))
      if cur_struct == value_struct and has_meta:
        value = meta.replace_boxed(cur, value)

    self.scope.put_variable(self.collection, self.name, value)

  def is_mutable(self) -> bool:
    """Checks if this Variable is mutable."""
    return self.scope.is_mutable_collection(self.collection)


class _ChildRNGSentinel:
  pass


# used to identify that an rng counter is meant for a child scope
child_rng_token = _ChildRNGSentinel()


class _DefaultSentinel:
  pass


# used to denote no default flag value on scope
no_flag = _DefaultSentinel()


class Scope:
  """A Scope allows easy access to variables and manages RNGS of a neural network layer.

  Scopes are purely functional and encapsulated in
  :class:`flax.linen.module.Module`, so users writing neural network code
  usually generally do not interact with ``Scopes`` directly.

  See `core design tests
  <https://github.com/google/flax/tree/main/tests/core/design>`_
  for a number of examples using ``Scopes``.
  """

  reservations: Dict[str, Set[Optional[str]]]

  def __init__(
    self,
    variables: MutableVariableDict,
    rngs: Optional[Union[RNGSequences, Dict[str, LazyRng]]] = None,
    name: Optional[str] = None,
    mutable: CollectionFilter = False,
    parent: Optional['Scope'] = None,
    path: Iterable[str] = (),
    debug_path: Iterable[str] = (),
    flags: Optional[Mapping] = None,
  ):
    """Initializes a Scope.

    Args:
      variables: VariableDict to initialize the Scope with.
      rngs: RNGs used in this scope or one of the child scopes.
      name: name of this scope.
      mutable: A CollectionFilter determining which variables are mutable.
      parent: The parent scope.
      path: The path in the variable tree from the root scope to this scope. It
        exactly matches the module path.
      debug_path: Similar to path but could contain transformation decorators.
      flags: internal flags.
    """
    rngs = {k: LazyRng.create(v) for k, v in rngs.items()} if rngs else {}
    self._variables = variables
    self.parent = parent
    self.name = name
    self.path = tuple(path)
    self.debug_path = tuple(debug_path) or self.path
    self.rngs = rngs
    self.mutable = mutable
    self.flags = freeze({} if flags is None else flags)

    self._root = parent.root if parent else None
    self.trace_level = tracers.trace_level(tracers.current_trace())

    self.rng_counters = {key: 0 for key in self.rngs}
    self.reservations = collections.defaultdict(set)

    self._invalid = False

  def __eq__(self, other: Any) -> bool:
    # If the root variable dict and path are the same, then two scopes behave
    # identically. Effectively, a scope is nothing more than a cursor into a
    # variable dict and an rng counter dict.
    if not isinstance(other, Scope):
      return False
    if self is other:
      return True
    return (
      self.root._variables is other.root._variables
      and self.path == other.path
      and self.rng_counters is other.rng_counters
    )

  def __hash__(self) -> int:
    # see __eq__
    return hash((id(self.root._variables), self.path, id(self.rng_counters)))

  @property
  def root(self) -> 'Scope':
    return self._root or self

  @property
  def path_text(self) -> str:
    """Returns the debug path as a human readable string."""
    return '/' + '/'.join(self.debug_path)

  @property
  def invalid(self) -> bool:
    """Returns true if this scope is invalidated as a result of `Scope.temporary`."""
    return self._invalid

  def _check_valid(self):
    if self._invalid:
      raise errors.InvalidScopeError(self.name)

  @contextlib.contextmanager
  def temporary(self):
    """Returns a context manager that will invalidate this Scope when leaving the context."""
    try:
      yield self
    finally:
      self.invalidate()

  def invalidate(self):
    """Invalidates the Scope."""
    self._invalid = True

  def mutable_variables(self) -> Union[VariableDict, Dict[str, Any]]:
    """Returns an immutable copy of the mutable variables belonging to this Scope."""
    self._populate_collections()
    xs = {
      k: v for k, v in self._variables.items() if in_filter(self.mutable, k)
    }
    if config.flax_return_frozendict:
      return freeze(xs)
    return xs

  def variables(self) -> Union[VariableDict, Dict[str, Any]]:
    """Returns an immutable copy of the variables belonging to this Scope."""
    self._populate_collections()
    if config.flax_return_frozendict:
      return freeze(self._variables)
    return self._variables

  def _validate_trace_level(self):
    tracers.check_trace_level(self.trace_level)

  def rewound(self, rewind_rngs: bool = False) -> 'Scope':
    """Returns a rewound version of this Scope.

    Args:
      rewind_rngs: if true, reset the RNG counter of this scope.

    Returns:
      A rewound version of this scope, which means reservations are
      emptied, and the rng counter is optionally rewound.
    """
    self._check_valid()
    scope = Scope(
      self._variables,
      self.rngs,
      self.name,
      self.mutable,
      self.parent,
      path=self.path,
      debug_path=self.debug_path,
      flags=self.flags,
    )
    if not rewind_rngs:
      scope.rng_counters = self.rng_counters
    return scope

  def name_reserved(self, name: str, col: Optional[str] = None) -> bool:
    """Checks whether a name for a child Scope or Variable is taken.

    Args:
      name: the name to check for collision.
      col: if a variable, the collection used.
    """
    if name in self.reservations:
      # allow the same name for two variables in
      # different collections, otherwise raise error.
      if (
        None in self.reservations[name]
        or col is None
        or col in self.reservations[name]
      ):
        return True
    return False

  def reserve(self, name: str, col: Optional[str] = None):
    """Reserves a name for a child Scope or Variable.

    Throws an error if the name exists already.

    Args:
      name: the name to reserve.
      col: if a variable, the collection used.
    """
    if not isinstance(name, str):
      raise TypeError(
        'The type of scope "{name}" should be string but ' f'it is {type(name)}'
      )
    if self.name_reserved(name, col):
      raise ValueError(f'Duplicate use of scope name: "{name}"')
    self.reservations[name].add(col)

  def default_name(self, prefix: str) -> str:
    """Generates an unreserved name with the given prefix.

    Args:
      prefix: prefix to use for generating an unreserved name.

    Returns:
      The generated name.
    """
    i = 0
    while True:
      name = f'{prefix}{i}'
      if name not in self.reservations:
        return name
      i += 1

  def push(
    self, name: Optional[str] = None, prefix: str = '', reuse=False
  ) -> 'Scope':
    """Creates a child Scope.

    Args:
      name: optional name of the child.
      prefix: prefix used for generating the name if `name` is `None`.
      reuse: if True will return a pre-existing child scope with the given name
        instead of throwing an error.

    Returns:
      The child scope.
    """
    self._check_valid()
    self._validate_trace_level()
    if name is None:
      name = self.default_name(prefix)
    if not reuse or name not in self.reservations:
      self.reserve(name)
    rngs = {key: LazyRng.create(rng, name) for key, rng in self.rngs.items()}
    rng_key = (child_rng_token, name)
    if rng_key in self.rng_counters:
      rng_counters = self.rng_counters.get(rng_key)  # type: ignore
    else:
      rng_counters = {key: 0 for key in rngs}
      self.rng_counters[rng_key] = rng_counters  # type: ignore
    scope = Scope(
      {},
      name=name,
      rngs=rngs,
      parent=self,
      mutable=self.mutable,
      path=self.path + (name,),
      debug_path=self.debug_path + (name,),
      flags=self.flags,
    )
    scope.rng_counters = rng_counters
    return scope

  def child(
    self,
    fn: Callable[..., Any],
    name: Optional[str] = None,
    prefix: Optional[str] = None,
    named_call: bool = True,
    **partial_kwargs,
  ) -> Callable[..., Any]:
    """Partially applies a child scope to fn.

    When calling the returned function multiple times variables will be reused.

    Args:
      fn: the function to partially apply the child Scope to.
      name: optional name of the child.
      prefix: prefix used for generating name if it is `None`.
      named_call: if true, `fn` will be run under `jax.named_scope`. The XLA
        profiler will use this to name tag the computation.
      **partial_kwargs: additional kwargs partially applied to `fn`.

    Returns:
      The function with a partially applied scope.
    """
    if name is None:
      if prefix is None:
        prefix = fn.__name__ + '_' if hasattr(fn, '__name__') else ''
      name = self.default_name(prefix)
    scope = self.push(name)

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
      kwargs = dict(partial_kwargs, **kwargs)
      if named_call:
        with jax.named_scope(name):
          res = fn(scope.rewound(), *args, **kwargs)
      else:
        res = fn(scope.rewound(), *args, **kwargs)
      return res

    return wrapper

  def is_mutable_collection(self, col: str) -> bool:
    """Returns true if the collection `col` is mutable."""
    return in_filter(self.mutable, col)

  def is_collection_empty(self, col: str) -> bool:
    """Returns true if the collection is empty."""
    if col in self.root._variables:  # pylint: disable=protected-access
      return not self.root._variables[col]  # pylint: disable=protected-access
    return True

  def _mutable_collection(self, col: str) -> MutableCollection:
    """Returns the collection `col` as a mutable object."""
    assert self.is_mutable_collection(col), f'Collection {col} is not mutable'

    # The actual variable dict is stored in the root scope only, and subscopes
    # hold references to subtrees relevant to them. This function ensures that
    # the collections are created in the top-level Scope and we return the
    # correct reference.
    if col not in self._variables:
      if not self.parent:
        # If this is the top-level Scope, just add an empty collection.
        self._variables[col] = {}
      else:
        assert self.name is not None  # Only top-level Scope have name None.
        # Populate the parent collections recursively and obtain a reference to
        # the direct parent (which, by transitivity, is be a reference to a
        # dict in the root Scope).
        parent_col = self.parent._mutable_collection(col)  # pylint: disable=protected-access
        if self.name not in parent_col:
          # If this Scope's name does not occur in the parent collection, add it
          # to the parent scope (updating the parent's variable dict).
          parent_col[self.name] = {}
        # Store a reference to the parent's scope collection for in this scope's
        # variable dict.
        self._variables[col] = parent_col[self.name]

    return self._variables[col]

  def _collection(self, col: str) -> Collection:
    """Returns a collection of variables of collection `col`."""
    if col not in self._variables:
      if self.parent:
        assert self.name is not None
        parent_col = self.parent._collection(col)  # pylint: disable=protected-access
        if self.name not in parent_col:
          return FrozenDict()
        self._variables[col] = parent_col[self.name]
      else:
        return FrozenDict()
    return self._variables[col]

  def has_rng(self, name: str) -> bool:
    """Returns true if a PRNGSequence with name `name` exists."""
    return name in self.rngs

  def make_rng(self, name: str = 'params') -> PRNGKey:
    """Generates A PRNGKey from a PRNGSequence with name `name`."""
    if not self.has_rng(name):
      if self.has_rng('params'):
        name = 'params'
      else:
        raise errors.InvalidRngError(f'{self.name} needs PRNG for "{name}"')
    self._check_valid()
    self._validate_trace_level()
    self.rng_counters[name] += 1
    return LazyRng.create(self.rngs[name], self.rng_counters[name]).as_jax_rng()

  def get_variable(self, col: str, name: str, default: Any = None) -> Any:
    """Retrieves the value of a Variable.

    Args:
      col: the variable collection.
      name: the name of the variable.
      default: the default value to return if the variable does not exist in
        this scope.

    Returns:
      The value of the input variable, of the default value if the variable
      doesn't exist in this scope.
    """
    variables = self._collection(col)
    if name in variables:
      return variables[name]
    else:
      return default

  def has_variable(self, col: str, name: str) -> bool:
    """Returns true if the given variable exists in this scope.

    Args:
      col: the collection of the variable.
      name: the name of the variable.
    """
    variables = self._collection(col)
    return name in variables

  def put_variable(self, col: str, name: str, value: Any):
    """Updates the value of the given variable if it is mutable, or an error otherwise.

    Args:
      col: the collection of the variable.
      name: the name of the variable.
      value: the new value of the given variable.
    """
    self._check_valid()
    self._validate_trace_level()
    if not self.is_mutable_collection(col):
      raise errors.ModifyScopeVariableError(col, name, self.path_text)
    variables = self._mutable_collection(col)

    # Make sure reference sharing of child variable dictionaries isn't broken.
    # See https://github.com/google/flax/issues/2022 for more details.
    def put(target, key, val):
      if (
        key in target
        and isinstance(target[key], dict)
        and isinstance(val, Mapping)
      ):
        for k, v in val.items():
          put(target[key], k, v)
      else:
        target[key] = val

    put(variables, name, value)

  @overload
  def variable(
    self,
    col: str,
    name: str,
    init_fn: Optional[Callable[..., T]] = None,
    *init_args,
  ) -> Variable[T]:
    ...

  @overload
  def variable(
    self,
    col: str,
    name: str,
    init_fn: Optional[Callable[..., T]] = None,
    *init_args,
    unbox: Literal[True],
    **init_kwargs,
  ) -> Variable[T]:
    ...

  @overload
  def variable(
    self,
    col: str,
    name: str,
    init_fn: Optional[Callable[..., T]] = None,
    *init_args,
    unbox: Literal[False],
    **init_kwargs,
  ) -> Variable[meta.AxisMetadata[T]]:
    ...

  @overload
  def variable(
    self,
    col: str,
    name: str,
    init_fn: Optional[Callable[..., T]] = None,
    *init_args,
    unbox: bool = True,
    **init_kwargs,
  ) -> Union[Variable[T], Variable[meta.AxisMetadata[T]]]:
    ...

  def variable(
    self,
    col: str,
    name: str,  # pylint: disable=keyword-arg-before-vararg
    init_fn: Optional[Callable[..., T]] = None,
    *init_args,
    unbox: bool = True,
    **init_kwargs,
  ) -> Union[Variable[T], Variable[meta.AxisMetadata[T]]]:
    """Creates a variable if it doesn't exist yet in this scope and returns it.

    Args:
      col: the collection of the variable.
      name: the name of the variable.
      init_fn: a function taking a PRNGKey plus any other number of positional
        arguments. If None, the variable must already be initialized otherwise
        an error is raised.
      *init_args: the positional arguments to evaluate init_fn on lazily.
      unbox: If True, ``AxisMetadata`` instances are replaced by their unboxed
        value, see ``flax.nn.meta.unbox`` (default: True).
      **init_kwargs: the key-word arguments to evaluate init_fn on lazily.

    Returns:
      The variable.  Throws an error if the variable exists already.
    """
    self.reserve(name, col)
    if not self.has_variable(col, name):
      if not self.is_mutable_collection(col) or init_fn is None:
        if self.is_collection_empty(col):
          raise errors.ScopeCollectionNotFound(col, name, self.path_text)
        raise errors.ScopeVariableNotFoundError(name, col, self.path_text)
      init_value = init_fn(*init_args, **init_kwargs)
      self.put_variable(col, name, init_value)
    # cast to make static analyzers happy
    return cast(
      Union[Variable[T], Variable[meta.AxisMetadata[T]]],
      Variable(self, col, name, unbox=unbox),
    )

  @overload
  def param(
    self, name: str, init_fn: Callable[..., T], *init_args,
  ) -> T:
    ...

  @overload
  def param(
    self,
    name: str,
    init_fn: Callable[..., T],
    *init_args,
    unbox: Literal[True],
    **init_kwargs,
  ) -> T:
    ...

  @overload
  def param(
    self,
    name: str,
    init_fn: Callable[..., T],
    *init_args,
    unbox: Literal[False],
    **init_kwargs,
  ) -> meta.AxisMetadata[T]:
    ...

  @overload
  def param(
    self,
    name: str,
    init_fn: Callable[..., T],
    *init_args,
    unbox: bool,
    **init_kwargs,
  ) -> Union[T, meta.AxisMetadata[T]]:
    ...

  def param(
    self,
    name: str,
    init_fn: Callable[..., T],
    *init_args,
    unbox: bool = True,
    **init_kwargs,
  ) -> Union[T, meta.AxisMetadata[T]]:
    """Creates a parameter if it doesn't exist yet in this scope and returns it.

    If the parameter exists already, the existing value is simply returned.

    Args:
      name: the name of the parameter.
      init_fn: a function taking a PRNGKey plus any other number of positional
        arguments.
      *init_args: the positional arguments to evaluate init_fn on lazily.
      unbox: If True, ``AxisMetadata`` instances are replaced by their unboxed
        value, see ``flax.nn.meta.unbox`` (default: True).
      **init_kwargs: the key-word arguments to evaluate init_fn on lazily.

    Returns:
      The parameters. Throws an error if the params exist already.
    """
    self.reserve(name, 'params')
    if self.has_variable('params', name):
      value = self.get_variable('params', name)
      # Validate that the shape of the init_fn output is the same as the shape
      # of the existing parameter. This is to make sure that the hparams set up
      # in a Flax Module match the shapes coming in during apply, and if not,
      # catch it with an error message.
      # NOTE: We could consider moving this to `self.`
      abs_value = jax.eval_shape(
        lambda: init_fn(random.key(0), *init_args, **init_kwargs)
      )
      abs_value_flat = jax.tree_util.tree_leaves(abs_value)
      value_flat = jax.tree_util.tree_leaves(value)
      for val, abs_val in zip(value_flat, abs_value_flat):
        # NOTE: We could check dtype consistency here as well but it's
        # usefuleness is less obvious. We might intentionally change the dtype
        # for inference to a half float type for example.
        if jnp.shape(val) != jnp.shape(abs_val):
          raise errors.ScopeParamShapeError(
            name, self.path_text, jnp.shape(abs_val), jnp.shape(val)
          )
    else:
      if not self.is_mutable_collection('params'):
        if self.is_collection_empty('params'):
          raise errors.ScopeCollectionNotFound('params', name, self.path_text)
        raise errors.ScopeParamNotFoundError(name, self.path_text)
      value = init_fn(self.make_rng('params'), *init_args, **init_kwargs)
      self.put_variable('params', name, value)
    if unbox:
      value = meta.unbox(value)
    return value

  def _populate_collections(self):
    collections = self.root._variables.keys()  # pylint: disable=protected-access
    for col in collections:
      self._collection(col)

  def has_flag(self, key) -> bool:
    return key in self.flags

  def get_flag(self, key, default=no_flag) -> Any:
    if key not in self.flags and default is no_flag:
      return ValueError(f'Flag {key} not present on scope.')
    return self.flags.get(key, default)


def _unfreeze_variables(variables, mutable):
  new_variables = {}
  for key, value in variables.items():
    if in_filter(mutable, key):
      new_variables[key] = unfreeze(value)
    else:
      new_variables[key] = value
  return new_variables


def bind(
  variables: VariableDict,
  rngs: Optional[RNGSequences] = None,
  mutable: CollectionFilter = False,
  flags: Optional[Mapping] = None,
):
  """Binds variables and rngs to a new ``Scope``.

  bind provides a ``Scope`` instance without transforming a function with
  ``apply``. This is particularly useful for debugging and interactive use cases
  like notebooks where a function would limit the ability split up code into
  different cells.

  a ``Scope`` instance is a stateful object. Note that idiomatic JAX is
  functional and therefore a ``Scope` does not mix well well with vanilla JAX
  APIs. Therefore, we recommend using ``apply`` when code should be reusable and
  compatible across the JAX software ecosystem.

  Args:
    variables: Variable dictionary to bind.
    rngs: RNGs to bind.
    mutable: Which variable collections to treat as mutable.
    flags: internal flags.

  Returns:
    A new scope with the variables and rngs bound to it.
  """
  if not _is_valid_variables(variables):
    raise errors.ApplyScopeInvalidVariablesTypeError()
  if rngs is not None and not _is_valid_rngs(rngs):
    raise errors.InvalidRngError(
      'rngs should be a dictionary mapping strings to `jax.PRNGKey`.'
    )
  new_variables = _unfreeze_variables(variables, mutable)
  return Scope(new_variables, rngs=rngs, mutable=mutable, flags=flags)


def apply(
  fn: Callable[..., Any],
  mutable: CollectionFilter = False,
  flags: Optional[Mapping] = None,
) -> Callable[..., Any]:
  """Functionalize a `Scope` function.

  Args:
    fn: a function taking a `Scope` as its first argument.
    mutable: the filter determining which variable collections are mutable.
    flags: internal flags.

  Returns:
    `fn` with the scope partially applied.
  """

  @functools.wraps(fn)
  def wrapper(
    variables: VariableDict,
    *args,
    rngs: Optional[Union[PRNGKey, RNGSequences]] = None,
    **kwargs,
  ) -> Union[Any, Tuple[Any, Union[VariableDict, Dict[str, Any]]]]:
    if rngs is not None:
      if not _is_valid_rng(rngs) and not _is_valid_rngs(rngs):
        raise ValueError(
          'The ``rngs`` argument passed to an apply function should be a '
          '``jax.PRNGKey`` or a dictionary mapping strings to '
          '``jax.PRNGKey``.'
        )
      if not isinstance(rngs, (dict, FrozenDict)):
        rngs = {'params': rngs}

    # Try to detect if user accidentally passed {'params': {'params': ...}.
    if (
      'params' in variables
      and isinstance(variables['params'], (dict, FrozenDict))
      and 'params' in variables['params']
    ):
      raise errors.ApplyScopeInvalidVariablesStructureError(variables)

    with bind(
      variables, rngs=rngs, mutable=mutable, flags=flags
    ).temporary() as root:
      y = fn(root, *args, **kwargs)
    if mutable is not False:
      return y, root.mutable_variables()
    else:
      return y

  return wrapper


def init(
  fn: Callable[..., Any],
  mutable: CollectionFilter = True,
  flags: Optional[Mapping] = None,
) -> Callable[..., Any]:
  """Functionalize a `Scope` function for initialization.

  Args:
    fn: a function taking a `Scope` as its first argument.
    mutable: the filter determining which variable collections are mutable.
    flags: internal flags.

  Returns:
    `fn` with the scope partially applied.
  """

  @functools.wraps(fn)
  def wrapper(rngs, *args, **kwargs) -> Tuple[Any, VariableDict]:
    if not _is_valid_rng(rngs) and not _is_valid_rngs(rngs):
      raise ValueError(
        'First argument passed to an init function should be a '
        '``jax.PRNGKey`` or a dictionary mapping strings to '
        '``jax.PRNGKey``.'
      )
    if not isinstance(rngs, (dict, FrozenDict)):
      rngs = {'params': rngs}
    init_flags = {**(flags if flags is not None else {}), 'initializing': True}
    return apply(fn, mutable=mutable, flags=init_flags)(
      {}, *args, rngs=rngs, **kwargs
    )

  return wrapper


def lazy_init(
  fn: Callable[..., Any],
  mutable: CollectionFilter = True,
  flags: Optional[Mapping] = None,
) -> Callable[..., Any]:
  """Functionalizes a `Scope` function for lazy initialization.

  Similair to ``init`` except that the init function now accepts
  ``jax.ShapeDtypeStruct`` instances for arguments that do not
  affect the variable initialization (typically this is all the input data).

  Example::

    def f(scope, x):
        # the kernel init only uses the shape of x so we don't actually
        # need a value for x and can pass it as a ShapeDtypeStruct in lazy_init.
        k = scope.param("kernel", nn.initializers.lecun_normal(), (x.shape[-1], x.shape[-1]))
        return x @ k
    init_fn = lazy_init(f)
    variables = init_fn(random.key(0), jax.ShapeDtypeStruct((1, 128), jnp.float32))


  Args:
    fn: a function taking a `Scope` as its first argument.
    mutable: the filter determining which variable collections are mutable.
    flags: internal flags.

  Returns:
    `fn` with the scope partially applied. Unlike ``init`` which returns a tuple of function
    output and variables, the lazy init function only returns the variables.
  """
  return partial_eval.lazy_init(
    lambda *args, **kwargs: init(fn, mutable, flags)(*args, **kwargs)[1]
  )


def _is_valid_collection(col: VariableDict):
  if not isinstance(col, (FrozenDict, dict)):
    return False
  for name in col.keys():
    # Any value can be stored in a collection so only keys can be verified.
    if not isinstance(name, str):
      return False
  return True


def _is_valid_variables(variables: VariableDict) -> bool:
  """Checks whether the given variable dict is valid.

  Args:
    variables: A variable dict.

  Returns:
    True if `variables` is a valid variable dict.
  """
  for name, col in variables.items():
    if not isinstance(name, str):
      return False
    if not _is_valid_collection(col):
      return False
  return True


def _is_valid_rng(rng: Array):
  """Checks whether rng is a valid JAX PRNGKey, also handling custom prngs."""
  # This check is valid for either new-style or old-style PRNG keys
  if not isinstance(rng, (np.ndarray, jnp.ndarray)):
    return False

  # Handle new-style typed PRNG keys
  if hasattr(jax.dtypes, 'prng_key'):  # JAX 0.4.14 or newer
    if jax.dtypes.issubdtype(rng.dtype, jax.dtypes.prng_key):
      return rng.shape == ()
  elif hasattr(jax.random, 'PRNGKeyArray'):  # Previous JAX versions
    if isinstance(rng, jax.random.PRNGKeyArray):
      return rng.shape == ()

  # Handle old-style raw PRNG keys
  expected_rng = jax.eval_shape(
    lambda s: jax.random.key_data(jax.random.key(s)), 0
  )
  if (rng.shape, rng.dtype) != (expected_rng.shape, expected_rng.dtype):
    return False
  return True


def _is_valid_rngs(rngs: Union[PRNGKey, RNGSequences]):
  if not isinstance(rngs, (FrozenDict, dict)):
    return False
  for key, val in rngs.items():
    if not isinstance(key, str):
      return False
    if not _is_valid_rng(val):
      return False
  return True



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
from typing import Sequence

import jax
import jax.numpy as jnp
from jax import lax, random

from flax.nnx.nnx import rnglib
from flax.nnx.nnx.module import Module, first_from


@dataclasses.dataclass
class Dropout(Module):
  """Create a dropout layer.

  To use dropout, call the :func:`train` method (or pass in
  ``deterministic=False`` in the constructor or during call time).

  To disable dropout, call the :func:`eval` method (or pass in
  ``deterministic=True`` in the constructor or during call time).

  Example usage::

    >>> from flax import nnx
    >>> import jax.numpy as jnp

    >>> class MLP(nnx.Module):
    ...   def __init__(self, rngs):
    ...     self.linear = nnx.Linear(in_features=3, out_features=4, rngs=rngs)
    ...     self.dropout = nnx.Dropout(0.5, rngs=rngs)
    ...   def __call__(self, x):
    ...     x = self.linear(x)
    ...     x = self.dropout(x)
    ...     return x

    >>> model = MLP(rngs=nnx.Rngs(0))
    >>> x = jnp.ones((1, 3))

    >>> model.train() # use dropout
    >>> model(x)
    Array([[-0.9353421,  0.       ,  1.434417 ,  0.       ]], dtype=float32)

    >>> model.eval() # don't use dropout
    >>> model(x)
    Array([[-0.46767104, -0.7213411 ,  0.7172085 , -0.31562346]], dtype=float32)

  Attributes:
    rate: the dropout probability.  (_not_ the keep rate!)
    broadcast_dims: dimensions that will share the same dropout mask
    deterministic: if false the inputs are scaled by ``1 / (1 - rate)`` and
      masked, whereas if true, no mask is applied and the inputs are returned
      as is.
    rng_collection: the rng collection name to use when requesting an rng key.
    rngs: rng key.
  """

  rate: float
  broadcast_dims: Sequence[int] = ()
  deterministic: bool = False
  rng_collection: str = 'dropout'
  rngs: rnglib.Rngs | None = None

  def __call__(
    self,
    inputs,
    *,
    deterministic: bool | None = None,
    rngs: rnglib.Rngs | None = None,
  ) -> jax.Array:
    """Applies a random dropout mask to the input.

    Args:
      inputs: the inputs that should be randomly masked.
      deterministic: if false the inputs are scaled by ``1 / (1 - rate)`` and
        masked, whereas if true, no mask is applied and the inputs are returned
        as is. The ``deterministic`` flag passed into the call method will take
        precedence over the ``deterministic`` flag passed into the constructor.
      rngs: rng key. The rng key passed into the call method will take
        precedence over the rng key passed into the constructor.

    Returns:
      The masked inputs reweighted to preserve mean.
    """
    deterministic = first_from(
      deterministic,
      self.deterministic,
      error_msg="""No `deterministic` argument was provided to Dropout
          as either a __call__ argument or class attribute""",
    )

    if (self.rate == 0.0) or deterministic:
      return inputs

    # Prevent gradient NaNs in 1.0 edge-case.
    if self.rate == 1.0:
      return jnp.zeros_like(inputs)

    rngs = first_from(
      rngs,
      self.rngs,
      error_msg="""`deterministic` is False, but no `rngs` argument was provided to Dropout
          as either a __call__ argument or class attribute.""",
    )

    keep_prob = 1.0 - self.rate
    rng = rngs[self.rng_collection]()
    broadcast_shape = list(inputs.shape)
    for dim in self.broadcast_dims:
      broadcast_shape[dim] = 1
    mask = random.bernoulli(rng, p=keep_prob, shape=broadcast_shape)
    mask = jnp.broadcast_to(mask, inputs.shape)
    return lax.select(mask, inputs / keep_prob, jnp.zeros_like(inputs))



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (
  Any,
  Callable,
  Dict,
  Generic,
  Hashable,
  Mapping,
  Optional,
  Protocol,
  Sequence,
  Tuple,
  TypeVar,
  Union,
  runtime_checkable,
)

import jax
from flax.core import FrozenDict

import dataclasses


# General

Array = Union[jax.Array, Any]
PRNGKey = jax.Array
RNGSequences = Dict[str, PRNGKey]
Dtype = Union[jax.typing.DTypeLike, Any]
Shape = Sequence[int]
K = TypeVar('K')

@runtime_checkable
class Key(Hashable, Protocol):
  def __lt__(self: K, value: K, /) -> bool:
    ...


Path = str
PathParts = Tuple[Key, ...]

Leaf = Any


# Linear

PrecisionLike = Union[
  None,
  str,
  jax.lax.Precision,
  Tuple[str, str],
  Tuple[jax.lax.Precision, jax.lax.Precision],
]
DotGeneralT = Callable[..., Array]
ConvGeneralDilatedT = Callable[..., Array]

PaddingLike = Union[str, int, Sequence[Union[int, Tuple[int, int]]]]
LaxPadding = Union[str, Sequence[Tuple[int, int]]]


# Initializers

Initializer = Union[jax.nn.initializers.Initializer, Callable[..., Any]]


# Collections

Collection = Mapping[str, Any]
MutableCollection = Dict[str, Any]


# Dicts

VariableDict = Mapping[str, Collection]
FrozenVariableDict = FrozenDict[str, Collection]
MutableVariableDict = Dict[str, MutableCollection]

PRNGFoldable = Union[int, str]


# Axes

T = TypeVar('T')

@dataclasses.dataclass(frozen=True)
class In(Generic[T]):
  """Specifies a variable collection should only be lifted as input."""

  axis: T

@dataclasses.dataclass(frozen=True)
class Out(Generic[T]):
  """Specifies a variable collection should only be lifted as output."""

  axis: T

Axis = Optional[int]
InOutAxis = Union[Axis, In[Axis], Out[Axis]]

ScanAxis = int
InOutScanAxis = Union[ScanAxis, In[ScanAxis], Out[ScanAxis]]

Axes = Union[int, Sequence[int]]


# SPMD

LogicalNames = Tuple[Union[str, None], ...]

# Maps each logical axis  to physical mesh, can be either None (replicated),
# one physical axis or a tuple of physical axes.
LogicalRules = Sequence[Tuple[str, Union[str, Tuple[str, ...], None]]]
ArrayPytree = Any  # pylint: disable=invalid-name
LogicalPartitionSpec = Any  # pylint: disable=invalid-name
LogicalPartitionSpecPytree = Any  # pylint: disable=invalid-name
PartitionSpecPytree = Any  # pylint: disable=invalid-name

Sharding = Tuple[Optional[str], ...]



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Input pipeline for VAE dataset."""

import jax
import jax.numpy as jnp
import tensorflow as tf
import tensorflow_datasets as tfds


def build_train_set(batch_size, ds_builder):
  """Builds train dataset."""

  train_ds = ds_builder.as_dataset(split=tfds.Split.TRAIN)
  train_ds = train_ds.map(prepare_image)
  train_ds = train_ds.cache()
  train_ds = train_ds.repeat()
  train_ds = train_ds.shuffle(50000)
  train_ds = train_ds.batch(batch_size)
  train_ds = iter(tfds.as_numpy(train_ds))
  return train_ds


def build_test_set(ds_builder):
  """Builds train dataset."""
  test_ds = ds_builder.as_dataset(split=tfds.Split.TEST)
  test_ds = test_ds.map(prepare_image).batch(10000)
  test_ds = jnp.array(list(test_ds)[0])
  test_ds = jax.device_put(test_ds)
  return test_ds


def prepare_image(x):
  x = tf.cast(x['image'], tf.float32)
  x = tf.reshape(x, (-1,))
  return x



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Batch apply."""

import jax, jax.numpy as jnp
import numpy as np


def ndim_at_least(x, num_dims):
  if not (isinstance(x, jax.Array) or isinstance(x, np.ndarray)):
    x = jnp.asarray(x)
  return x.ndim >= num_dims

def arbitrary_mergeable_leaf(min_num_dims, args, kwargs):
  for a in jax.tree_util.tree_leaves(args):
    if ndim_at_least(a, min_num_dims):
      return a
  for k in jax.tree_util.tree_leaves(kwargs):
    if ndim_at_least(k, min_num_dims):
      return k
  # Couldn't find a satisfactory leaf.
  return None

def merge_leading_dims(x, num_dims):
  """Merge leading dimensions."""
  # Don't merge if there aren't dimensions to merge.
  if not ndim_at_least(x, num_dims):
    return x

  new_shape = (np.prod(x.shape[:num_dims]),) + x.shape[num_dims:]
  return x.reshape(new_shape)

def split_leading_dim(x, to_dim):
  new_shape = to_dim + x.shape[1:]
  return x.reshape(new_shape)

class BatchApply:
  r"""Temporarily merges leading dimensions of input tensors.

  Merges the leading dimensions of a tensor into a single dimension, runs the
  given callable, then splits the leading dimension of the result to match the
  input.

  Input arrays whose rank is smaller than the number of dimensions to collapse
  are passed unmodified.

  This may be useful for applying a module to each timestep of e.g. a
  ``[Time, Batch, ...]`` array.

  For some ``f``\ s and platforms, this may be more efficient than
  :func:`jax.vmap`, especially when combined with other transformations like
  :func:`jax.grad`.

  Example usage::

    >>> import jax, jax.numpy as jnp

    >>> a = jax.random.normal(jax.random.key(0), [2, 3, 4])
    >>> b = jax.random.normal(jax.random.key(1), [4])

    >>> def raises(a, b):
    ...   if len(a.shape) != 2:
    ...     raise ValueError("a must be shape 2")
    ...   if len(b.shape) != 1:
    ...     raise ValueError("b must be shape 1")
    ...   return jnp.dot(a, b)

    >>> out = BatchApply(raises)(a, b)
    >>> expected_merged_leading = raises(a.reshape(2*3, 4), b)
    >>> expected = expected_merged_leading.reshape((2, 3) + expected_merged_leading.shape[1:])
    >>> np.testing.assert_array_equal(out, expected)
  """

  def __init__(self, f, num_dims=2):
    """Constructs a :class:`BatchApply` module.

    Args:
      f: The callable to be applied to the reshaped array.
      num_dims: The number of dimensions to merge.
    """
    self._f = f
    self.num_dims = num_dims

  def __call__(self, *args, **kwargs):
    example = arbitrary_mergeable_leaf(self.num_dims, args, kwargs)
    if example is None:
      raise ValueError(
        'BatchApply requires at least one input with ndim >= '
        f'{self.num_dims}.'
      )

    merge = lambda x: merge_leading_dims(x, self.num_dims)
    split = lambda x: split_leading_dim(x, example.shape[:self.num_dims])
    args = jax.tree_util.tree_map(merge, args)
    kwargs = jax.tree_util.tree_map(merge, kwargs)
    outputs = self._f(*args, **kwargs)
    return jax.tree_util.tree_map(split, outputs)


# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class and functions to define and initialize the actor-critic model."""

from flax import linen as nn
import jax.numpy as jnp


class ActorCritic(nn.Module):
  """Class defining the actor-critic model."""

  num_outputs: int

  @nn.compact
  def __call__(self, x):
    """Define the convolutional network architecture.

    Architecture originates from "Human-level control through deep reinforcement
    learning.", Nature 518, no. 7540 (2015): 529-533.
    Note that this is different than the one from  "Playing atari with deep
    reinforcement learning." arxiv.org/abs/1312.5602 (2013)

    Network is used to both estimate policy (logits) and expected state value;
    in other words, hidden layers' params are shared between policy and value
    networks, see e.g.:
    github.com/openai/baselines/blob/master/baselines/ppo1/cnn_policy.py
    """
    dtype = jnp.float32
    x = x.astype(dtype) / 255.0
    x = nn.Conv(
        features=32,
        kernel_size=(8, 8),
        strides=(4, 4),
        name='conv1',
        dtype=dtype,
    )(x)
    x = nn.relu(x)
    x = nn.Conv(
        features=64,
        kernel_size=(4, 4),
        strides=(2, 2),
        name='conv2',
        dtype=dtype,
    )(x)
    x = nn.relu(x)
    x = nn.Conv(
        features=64,
        kernel_size=(3, 3),
        strides=(1, 1),
        name='conv3',
        dtype=dtype,
    )(x)
    x = nn.relu(x)
    x = x.reshape((x.shape[0], -1))  # flatten
    x = nn.Dense(features=512, name='hidden', dtype=dtype)(x)
    x = nn.relu(x)
    logits = nn.Dense(features=self.num_outputs, name='logits', dtype=dtype)(x)
    policy_log_probabilities = nn.log_softmax(logits)
    value = nn.Dense(features=1, name='value', dtype=dtype)(x)
    return policy_log_probabilities, value



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import inspect
import typing as tp

import jax
import jax.numpy as jnp
import optax

from flax.nnx.nnx.graph import Key
from flax.nnx.nnx.module import GraphDef, Module
from flax.nnx.nnx.proxy_caller import ApplyCaller
from flax.nnx.nnx.rnglib import Rngs
from flax.nnx.nnx.state import State
from flax.training.train_state import struct

A = tp.TypeVar('A')
M = tp.TypeVar('M', bound=Module)
TS = tp.TypeVar('TS', bound='TrainState')


class Dict(Module, tp.Mapping[str, A]):
  @tp.overload
  def __init__(self, iterable: tp.Iterable[tp.Tuple[str, A]], /): ...

  @tp.overload
  def __init__(
    self, mapping: tp.Optional[tp.Mapping[str, A]] = None, /, **kwargs: A
  ): ...

  def __init__(self, *args, **kwargs):
    for name, value in dict(*args, **kwargs).items():
      setattr(self, name, value)

  def __getitem__(self, key) -> A:
    return getattr(self, key)

  def __setitem__(self, key, value):
    setattr(self, key, value)

  def __getattr__(self, key) -> A:
    return super().__getattribute__(key)

  def __setattr__(self, key, value):
    super().__setattr__(key, value)

  def __iter__(self) -> tp.Iterator[str]:
    return (k for k in vars(self) if k != '_object__state')

  def __len__(self) -> int:
    return len(vars(self))


class List(Module, tp.Generic[A]):
  def __init__(self, elems: tp.Iterable[A], /):
    i = 0
    for i, value in enumerate(elems):
      setattr(self, str(i), value)
    self._length = i + 1

  def __getitem__(self, key: int) -> A:
    if key >= len(self) or key < -len(self):
      raise IndexError(f'index {key} out of range for {self}')
    if key < 0:
      key = self._length + key
    return getattr(self, str(key))

  def __setitem__(self, key: int, value: A):
    if key >= len(self):
      raise IndexError(f'index {key} out of range for {self}')
    setattr(self, str(key), value)

  def __iter__(self) -> tp.Iterator[A]:
    for i in range(len(self)):
      yield getattr(self, str(i))

  def __len__(self) -> int:
    return self._length

  def _graph_node_flatten(self):
    nodes: list[tuple[Key, tp.Any]] = sorted(
      (int(key), value)
      for key, value in vars(self).items()
      if key not in ('_object__state', '_length')
    )
    nodes.append(('_length', self._length))
    return nodes, type(self)

  def _graph_node_set_key(self, key: Key, value: tp.Any):
    if isinstance(key, int):
      key = str(key)
    return super()._graph_node_set_key(key, value)

  def _graph_node_pop_key(self, key: Key):
    if isinstance(key, int):
      key = str(key)
    return super()._graph_node_pop_key(key)


class Sequential(Module):
  def __init__(self, *fns: tp.Callable[..., tp.Any]):
    self.layers = list(fns)

  def __call__(self, *args, rngs: tp.Optional[Rngs] = None, **kwargs) -> tp.Any:
    output: tp.Any = None

    for i, f in enumerate(self.layers):
      if not callable(f):
        raise TypeError(f'Sequence[{i}] is not callable: {f}')
      if i > 0:
        if isinstance(output, tuple):
          args = output
          kwargs = {}
        elif isinstance(output, dict):
          args = ()
          kwargs = output
        else:
          args = (output,)
          kwargs = {}
      if rngs is not None and has_keyword_arg(f, 'rngs'):
        kwargs['rngs'] = rngs

      output = f(*args, **kwargs)

    return output


class ModuleDefApply(tp.Protocol, tp.Generic[M]):
  def __call__(
    self, state: State, *states: State
  ) -> ApplyCaller[tuple[State, GraphDef[M]]]: ...


class TrainState(tp.Generic[M], struct.PyTreeNode):
  graphdef: GraphDef[M]
  params: State
  opt_state: optax.OptState
  step: jax.Array
  tx: optax.GradientTransformation = struct.field(pytree_node=False)

  @classmethod
  def create(
    cls,
    graphdef: GraphDef[M],
    *,
    params: State,
    tx: optax.GradientTransformation,
    step: int = 0,
    **kwargs,
  ):
    return cls(
      graphdef=graphdef,
      params=params,
      opt_state=tx.init(params),
      step=jnp.asarray(step),
      tx=tx,
      **kwargs,
    )

  if tp.TYPE_CHECKING:

    def __getattr__(self, key: str) -> tp.Any: ...

  def apply(
    self, state: tp.Union[State, str], *states: tp.Union[State, str]
  ) -> ApplyCaller[tuple[GraphDef[M], State]]:
    states = (state, *states)

    _states: list[State] = []

    for _state in states:
      if isinstance(_state, str):
        _state_key = _state
        _state = getattr(self, _state_key)
        if not isinstance(_state, State):
          raise TypeError(
            f'Expected {self.__class__.__name__}.{_state_key} to be a State, got {type(_state)}'
          )
      _states.append(_state)

    return self.graphdef.apply(*_states)

  def apply_gradients(self: TS, grads: State, **kwargs) -> TS:
    updates, opt_state = self.tx.update(grads, self.opt_state, self.params)
    params = optax.apply_updates(self.params, updates)  # type: ignore
    step = self.step + 1
    return self.replace(
      params=params,
      opt_state=opt_state,
      step=step,
      **kwargs,
    )


def has_keyword_arg(func: tp.Callable[..., tp.Any], name: str) -> bool:
  """Return True if func has keyword-only arguments with the given name."""
  return any(
    param.name == name
    and param.kind in (param.KEYWORD_ONLY, param.POSITIONAL_OR_KEYWORD)
    for param in inspect.signature(func).parameters.values()
  )



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.struct and flax.serialization."""

import collections
import platform
from typing import Any, NamedTuple

import jax
import jax.numpy as jnp
import msgpack
import numpy as np
import optax
import pytest
from absl.testing import absltest, parameterized
from jax import random
from jax.tree_util import Partial

from flax import linen as nn
from flax import serialization, struct
from flax.core import freeze
from flax.training import train_state

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


@struct.dataclass
class Point:
  x: float
  y: float
  meta: Any = struct.field(pytree_node=False)


@struct.dataclass
class Box:
  value: int


def to_state_dict(box: Box):
  return box.value


def from_state_dict(box: Box, state: Any):
  return box.replace(value=state)


serialization.register_serialization_state(
  Box, to_state_dict, from_state_dict, override=True
)


class OriginalTuple(NamedTuple):
  value: Any


class WrongTuple(NamedTuple):
  wrong_field: Any


class OriginalModule(nn.Module):
  @nn.compact
  def __call__(self, x):
    x = nn.Dense(10)(x)
    return x


class WrongModule(nn.Module):
  @nn.compact
  def __call__(self, x):
    x = nn.Dense(10)(x)
    x = nn.Dense(10)(x)
    return x


class SerializationTest(parameterized.TestCase):
  def test_dataclass_serialization(self):
    p = Point(x=1, y=2, meta={'dummy': True})
    state_dict = serialization.to_state_dict(p)
    self.assertEqual(
      state_dict,
      {
        'x': 1,
        'y': 2,
      },
    )
    restored_p = serialization.from_state_dict(p, {'x': 3, 'y': 4})
    expected_p = Point(x=3, y=4, meta={'dummy': True})
    self.assertEqual(restored_p, expected_p)

    with self.assertRaises(ValueError):  # invalid field
      serialization.from_state_dict(p, {'z': 3})
    with self.assertRaises(ValueError):  # missing field
      serialization.from_state_dict(p, {'x': 3})

  def test_pass_through_serialization(self):
    p = Box(value=123)
    state_dict = serialization.to_state_dict(p)
    self.assertEqual(
      state_dict,
      123,
    )
    restored_box = serialization.from_state_dict(p, 123)
    expected_box = Box(value=123)
    self.assertEqual(restored_box, expected_box)

  def test_model_serialization(self):
    rng = random.key(0)
    module = nn.Dense(features=1, kernel_init=nn.initializers.ones_init())
    x = jnp.ones((1, 1), jnp.float32)
    initial_params = module.init(rng, x)
    state = serialization.to_state_dict(initial_params)
    self.assertEqual(
      state,
      {
        'params': {
          'kernel': np.ones((1, 1)),
          'bias': np.zeros((1,)),
        }
      },
    )
    state = {
      'params': {
        'kernel': np.zeros((1, 1)),
        'bias': np.zeros((1,)),
      }
    }
    restored_model = serialization.from_state_dict(initial_params, state)
    self.assertEqual(restored_model, freeze(state))

  def test_partial_serialization(self):
    add_one = Partial(jnp.add, 1)
    state = serialization.to_state_dict(add_one)
    self.assertEqual(state, {'args': {'0': 1}, 'keywords': {}})
    restored_add_one = serialization.from_state_dict(add_one, state)
    self.assertEqual(add_one.args, restored_add_one.args)

  def test_optimizer_serialization(self):
    rng = random.key(0)
    module = nn.Dense(features=1, kernel_init=nn.initializers.ones_init())
    x = jnp.ones((1, 1), jnp.float32)
    initial_params = module.init(rng, x)
    tx = optax.sgd(0.1, momentum=0.1)
    tx_state = tx.init(initial_params)
    state = serialization.to_state_dict(tx_state)
    expected_state = {
      '0': {
        'trace': {
          'params': {
            'bias': np.array([0.0], dtype=jnp.float32),
            'kernel': np.array([[0.0]], dtype=jnp.float32),
          }
        }
      },
      '1': {},
    }
    self.assertEqual(state, expected_state)
    state = jax.tree_util.tree_map(lambda x: x + 1, expected_state)
    restored_tx_state = serialization.from_state_dict(tx_state, state)
    tx_state_plus1 = jax.tree_util.tree_map(lambda x: x + 1, tx_state)
    self.assertEqual(restored_tx_state, tx_state_plus1)

  def test_collection_serialization(self):
    @struct.dataclass
    class DummyDataClass:
      x: float

      @classmethod
      def initializer(cls, shape):
        del shape
        return cls(x=0.0)

    class StatefulModule(nn.Module):
      @nn.compact
      def __call__(self):
        state = self.variable('state', 'dummy', DummyDataClass.initializer, ())
        state.value = state.value.replace(x=state.value.x + 1.0)

    initial_variables = StatefulModule().init(random.key(0))
    _, variables = StatefulModule().apply(initial_variables, mutable=['state'])
    serialized_state_dict = serialization.to_state_dict(variables)
    self.assertEqual(serialized_state_dict, {'state': {'dummy': {'x': 2.0}}})
    deserialized_state = serialization.from_state_dict(
      variables, serialized_state_dict
    )
    self.assertEqual(variables, deserialized_state)

  @parameterized.parameters(
    [
      'byte',
      'b',
      'ubyte',
      'short',
      'h',
      'ushort',
      'i',
      'uint',
      'intp',
      'p',
      'uintp',
      'long',
      'l',
      'longlong',
      'q',
      'ulonglong',
      'half',
      'e',
      'f',
      'double',
      'd',
      'longdouble',
      'g',
      'cfloat',
      'cdouble',
      'clongdouble',
      'm',
      'bool8',
      'b1',
      'int64',
      'i8',
      'uint64',
      'u8',
      'float16',
      'f2',
      'float32',
      'f4',
      'float64',
      'f8',
      'float128',
      'f16',
      'complex64',
      'c8',
      'complex128',
      'c16',
      'complex256',
      'c32',
      'm8',
      'int32',
      'i4',
      'uint32',
      'u4',
      'int16',
      'i2',
      'uint16',
      'u2',
      'int8',
      'i1',
      'uint8',
      'u1',
      'complex_',
      'int0',
      'uint0',
      'single',
      'csingle',
      'singlecomplex',
      'float_',
      'intc',
      'uintc',
      'int_',
      'longfloat',
      'clongfloat',
      'longcomplex',
      'bool_',
      'int',
      'float',
      'complex',
      'bool',
    ]
  )
  def test_numpy_serialization(self, dtype):
    np.random.seed(0)
    if (
      (dtype in {'float128', 'f16', 'complex256', 'c32'})
      and (platform.system() == 'Darwin')
      and (platform.machine() == 'arm64')
    ):
      pytest.skip(
        f'Mac M1 does not support dtype {dtype}'
      )  # skip testing these dtypes if user is on Mac M1

    v = np.random.uniform(-100, 100, size=()).astype(dtype)[()]
    restored_v = serialization.msgpack_restore(
      serialization.msgpack_serialize(v)
    )
    self.assertEqual(restored_v.dtype, v.dtype)
    np.testing.assert_array_equal(restored_v, v)

    for shape in [(), (5,), (10, 10), (1, 20, 30, 1)]:
      arr = np.random.uniform(-100, 100, size=shape).astype(dtype)
      restored_arr = serialization.msgpack_restore(
        serialization.msgpack_serialize(arr)
      )
      self.assertEqual(restored_arr.dtype, arr.dtype)
      np.testing.assert_array_equal(restored_arr, arr)

  def test_jax_numpy_serialization(self):
    jax_dtypes = [
      jnp.bool_,
      jnp.uint8,
      jnp.uint16,
      jnp.uint32,
      jnp.int8,
      jnp.int16,
      jnp.int32,
      jnp.bfloat16,
      jnp.float16,
      jnp.float32,
      jnp.complex64,
    ]
    for dtype in jax_dtypes:
      v = jnp.array(np.random.uniform(-100, 100, size=())).astype(dtype)[()]
      restored_v = serialization.msgpack_restore(
        serialization.msgpack_serialize(v)
      )
      self.assertEqual(restored_v.dtype, v.dtype)
      np.testing.assert_array_equal(restored_v, v)

      for shape in [(), (5,), (10, 10), (1, 20, 30, 1)]:
        arr = jnp.array(np.random.uniform(-100, 100, size=shape)).astype(dtype)
        restored_arr = serialization.msgpack_restore(
          serialization.msgpack_serialize(arr)
        )
        self.assertEqual(restored_arr.dtype, arr.dtype)
        np.testing.assert_array_equal(restored_arr, arr)

  def test_complex_serialization(self):
    for x in [1j, 1 + 2j]:
      restored_x = serialization.msgpack_restore(
        serialization.msgpack_serialize(x)
      )
      self.assertEqual(x, restored_x)

  def test_restore_chunked(self):
    old_chunksize = serialization.MAX_CHUNK_SIZE
    serialization.MAX_CHUNK_SIZE = 91 * 8
    try:
      tmp = np.random.uniform(-100, 100, size=(21, 37))
      serialized = serialization.to_bytes(tmp)
      restored = serialization.msgpack_restore(serialized)
    finally:
      serialization.MAX_CHUNK_SIZE = old_chunksize

    np.testing.assert_array_equal(restored, tmp)

  def test_restore_unchunked(self):
    """Check if mgspack_restore works for unchunked inputs."""

    def msgpack_serialize_legacy(pytree):
      """Old implementation that was not chunking."""
      return msgpack.packb(
        pytree, default=serialization._msgpack_ext_pack, strict_types=True
      )

    tmp = np.random.uniform(-100, 100, size=(21, 37))
    serialized = msgpack_serialize_legacy(tmp)
    old_chunksize = serialization.MAX_CHUNK_SIZE
    serialization.MAX_CHUNK_SIZE = 91 * 8
    try:
      restored = serialization.msgpack_restore(serialized)
    finally:
      serialization.MAX_CHUNK_SIZE = old_chunksize

    np.testing.assert_array_equal(restored, tmp)

  def test_namedtuple_serialization(self):
    foo_class = collections.namedtuple('Foo', 'a b c')
    x1 = foo_class(a=1, b=2, c=3)
    x1_serialized = serialization.to_bytes(x1)
    x2 = foo_class(a=0, b=0, c=0)
    restored_x1 = serialization.from_bytes(x2, x1_serialized)
    self.assertEqual(type(x1), type(restored_x1))
    self.assertEqual(x1, restored_x1)

  def test_namedtuple_restore_legacy(self):
    foo_class = collections.namedtuple('Foo', 'a b c')
    x1 = foo_class(a=1, b=2, c=3)
    legacy_encoding = {
      'name': 'Foo',
      'fields': {'0': 'a', '1': 'b', '2': 'c'},
      'values': {'0': 1, '1': 2, '2': 3},
    }
    x2 = foo_class(a=0, b=0, c=0)
    restored_x1 = serialization.from_state_dict(x2, legacy_encoding)
    self.assertEqual(type(x1), type(restored_x1))
    self.assertEqual(x1, restored_x1)

  def test_model_serialization_to_bytes(self):
    rng = random.key(0)
    module = nn.Dense(features=1, kernel_init=nn.initializers.ones_init())
    initial_params = module.init(rng, jnp.ones((1, 1), jnp.float32))
    serialized_bytes = serialization.to_bytes(initial_params)
    restored_params = serialization.from_bytes(initial_params, serialized_bytes)
    self.assertEqual(restored_params, initial_params)

  def test_optimizer_serialization_to_bytes(self):
    rng = random.key(0)
    module = nn.Dense(features=1, kernel_init=nn.initializers.ones_init())
    initial_params = module.init(rng, jnp.ones((1, 1), jnp.float32))
    # model = nn.Model(module, initial_params)
    tx = optax.sgd(0.1, momentum=0.1)
    tx_state = tx.init(initial_params)
    serialized_bytes = serialization.to_bytes(tx_state)
    restored_tx_state = serialization.from_bytes(tx_state, serialized_bytes)
    self.assertEqual(restored_tx_state, tx_state)

  def test_serialization_chunking(self):
    old_chunksize = serialization.MAX_CHUNK_SIZE
    serialization.MAX_CHUNK_SIZE = 91 * 8
    try:
      tmp = {'a': np.ones((10, 10))}
      tmp = serialization._chunk_array_leaves_in_place(tmp)
    finally:
      serialization.MAX_CHUNK_SIZE = old_chunksize
    test = jax.tree_util.tree_map(jnp.shape, tmp)
    ref = {
      'a': {
        '__msgpack_chunked_array__': (),
        'chunks': {'0': (91,), '1': (9,)},
        'shape': {'0': (), '1': ()},
      }
    }
    self.assertEqual(test, ref)

  def test_serialization_chunking2(self):
    old_chunksize = serialization.MAX_CHUNK_SIZE
    serialization.MAX_CHUNK_SIZE = 91 * 8
    try:
      tmp = {'a': np.ones((10, 10))}
      tmpbytes = serialization.to_bytes(tmp)
      newtmp = serialization.from_bytes(tmp, tmpbytes)
    finally:
      serialization.MAX_CHUNK_SIZE = old_chunksize
    jax.tree_util.tree_map(np.testing.assert_array_equal, tmp, newtmp)

  def test_serialization_chunking3(self):
    old_chunksize = serialization.MAX_CHUNK_SIZE
    serialization.MAX_CHUNK_SIZE = 91 * 8
    try:
      tmp = {'a': np.ones((10, 10))}
      tmpbytes = serialization.msgpack_serialize(tmp)
      newtmp = serialization.msgpack_restore(tmpbytes)
    finally:
      serialization.MAX_CHUNK_SIZE = old_chunksize

    jax.tree_util.tree_map(np.testing.assert_array_equal, tmp, newtmp)

  @parameterized.parameters(
    {
      'target': [[[1, 2, 3], [4, 5]]],
      'wrong_target': [[[1, 2, 3], [4]]],
      'msg': (
        'The size of the list and the state dict do not match,'
        ' got 1 and 2 at path ./0/1'
      ),
    },
    {
      'target': (((1, 2, 3), (4, 5)),),
      'wrong_target': (((1, 2, 3), (4,)),),
      'msg': (
        'The size of the list and the state dict do not match,'
        ' got 1 and 2 at path ./0/1'
      ),
    },
    {
      'target': (((1, 2, 3), (OriginalTuple([4, 5]), 6)),),
      'wrong_target': (((1, 2, 3), (WrongTuple([4, 5]), 6)),),
      'msg': (
        'The field names of the state dict and the named tuple do '
        "not match, got {'value'} and {'wrong_field'} at path ./0/1/0"
      ),
    },
    {
      'target': {'a': {'b': {'c': [1, 2, 3], 'd': [4, 5]}}},
      'wrong_target': {'a': {'b': {'c': [1, 2, 3], 'd': [4]}}},
      'msg': (
        'The size of the list and the state dict do not match,'
        ' got 1 and 2 at path ./a/b/d'
      ),
    },
    {
      'target': {'a': {'b': {'c': [1, 2, 3], 'd': [4, 5]}}},
      'wrong_target': {'a': {'b': {'c': [1, 2, 3], 'e': [4, 5]}}},
      'msg': (
        'The target dict keys and state dict keys do not match, target'
        " dict contains keys {'e'} which are not present in state dict at"
        ' path ./a/b'
      ),
    },
    {
      'target': 'original_params',
      'wrong_target': 'wrong_params',
      'msg': (
        'The target dict keys and state dict keys do not match, target'
        " dict contains keys {'Dense_1'} which are not present in state"
        ' dict at path ./params'
      ),
    },
    {
      'target': 'original_train_state',
      'wrong_target': 'wrong_train_state',
      'msg': (
        'The target dict keys and state dict keys do not match, target'
        " dict contains keys {'Dense_1'} which are not present in state"
        ' dict at path ./params/params'
      ),
    },
  )
  def test_serialization_errors(self, target, wrong_target, msg):
    if target == 'original_params':
      x = jnp.ones((1, 28, 28, 1))
      rng = jax.random.key(1)
      original_module = OriginalModule()
      target = original_module.init(rng, x)
      wrong_module = WrongModule()
      wrong_target = wrong_module.init(rng, x)

    elif target == 'original_train_state':
      x = jnp.ones((1, 28, 28, 1))
      rng = jax.random.key(1)
      original_module = OriginalModule()
      original_params = original_module.init(rng, x)
      wrong_module = WrongModule()
      wrong_params = wrong_module.init(rng, x)

      tx = optax.sgd(learning_rate=0.1, momentum=0.9)
      target = train_state.TrainState.create(
        apply_fn=original_module.apply, params=original_params, tx=tx
      )
      wrong_target = train_state.TrainState.create(
        apply_fn=wrong_module.apply, params=wrong_params, tx=tx
      )

    encoded_bytes = serialization.to_bytes(target)
    with self.assertRaisesWithLiteralMatch(ValueError, msg):
      serialization.from_bytes(wrong_target, encoded_bytes)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Write Summaries from JAX for use with Tensorboard."""

import contextlib
import functools
import os

import numpy as np
import tensorflow as tf  # pytype: disable=import-error
from tensorboard.plugins.hparams import api as hparams_api

# pylint: disable=g-import-not-at-top
from flax import io


def _flatten_dict(input_dict, parent_key='', sep='.'):
  """Flattens and simplifies dict such that it can be used by hparams.

  Args:
    input_dict: Input dict, e.g., from ConfigDict.
    parent_key: String used in recursion.
    sep: String used to separate parent and child keys.

  Returns:
   Flattened dict.
  """
  items = []
  for k, v in input_dict.items():
    new_key = parent_key + sep + k if parent_key else k

    # Valid types according to https://github.com/tensorflow/tensorboard/blob/1204566da5437af55109f7a4af18f9f8b7c4f864/tensorboard/plugins/hparams/summary_v2.py
    valid_types = (
      bool,
      int,
      float,
      str,
      np.bool_,
      np.integer,
      np.floating,
      np.character,
    )

    if isinstance(v, dict):
      # Recursively flatten the dict.
      items.extend(_flatten_dict(v, new_key, sep=sep).items())
      continue
    elif not isinstance(v, valid_types):
      # Cast any incompatible values as strings such that they can be handled by hparams
      v = str(v)
    items.append((new_key, v))
  return dict(items)


@contextlib.contextmanager
def _as_default(summary_writer: tf.summary.SummaryWriter, auto_flush: bool):
  """No-flush variation of summary_writer.as_default()."""
  context_manager = summary_writer.as_default()
  try:
    context_manager.__enter__()
    yield summary_writer
  finally:
    old_flush = summary_writer.flush
    new_flush = old_flush if auto_flush else lambda: None
    summary_writer.flush = new_flush
    context_manager.__exit__(None, None, None)
    summary_writer.flush = old_flush


class SummaryWriter:
  """Saves data in event and summary protos for tensorboard."""

  def __init__(self, log_dir, auto_flush=True):
    """Create a new SummaryWriter.

    Args:
      log_dir: path to record tfevents files in.
      auto_flush: if true, flush after every reported metric.
    """
    log_dir = os.fspath(log_dir)

    # If needed, create log_dir directory as well as missing parent directories.
    if not io.isdir(log_dir):
      io.makedirs(log_dir)

    self._event_writer = tf.summary.create_file_writer(log_dir)
    self._as_default = functools.partial(_as_default, auto_flush=auto_flush)
    self._closed = False

  def close(self):
    """Close SummaryWriter. Final!"""
    if not self._closed:
      self._event_writer.close()
      self._closed = True
      del self._event_writer

  def flush(self):
    self._event_writer.flush()

  def scalar(self, tag, value, step):
    """Saves scalar value.

    Args:
      tag: str: label for this data
      value: int/float: number to log
      step: int: training step
    """
    value = float(np.array(value))
    with self._as_default(self._event_writer):
      tf.summary.scalar(name=tag, data=value, step=step)

  def image(self, tag, image, step, max_outputs=3):
    """Saves RGB image summary from np.ndarray [H,W], [H,W,1], or [H,W,3].

    Args:
      tag: str: label for this data
      image: ndarray: [H,W], [H,W,1], [H,W,3], [K,H,W], [K,H,W,1], [K,H,W,3]
        Save image in greyscale or colors.
        Pixel values could be either uint8 or float.
        Floating point values should be in range [0, 1).
      step: int: training step
      max_outputs: At most this many images will be emitted at each step.
        Defaults to 3.
    """
    image = np.array(image)
    # tf.summary.image expects image to have shape [k, h, w, c] where,
    # k = number of samples, h = height, w = width, c = number of channels.
    if len(np.shape(image)) == 2:
      image = image[np.newaxis, :, :, np.newaxis]
    elif len(np.shape(image)) == 3:
      # this could be either [k, h, w] or [h, w, c]
      if np.shape(image)[-1] in (1, 3):
        image = image[np.newaxis, :, :, :]
      else:
        image = image[:, :, :, np.newaxis]
    if np.shape(image)[-1] == 1:
      image = np.repeat(image, 3, axis=-1)

    # Convert to tensor value as tf.summary.image expects data to be a tensor.
    image = tf.convert_to_tensor(image)
    with self._as_default(self._event_writer):
      tf.summary.image(name=tag, data=image, step=step, max_outputs=max_outputs)

  def audio(self, tag, audiodata, step, sample_rate=44100, max_outputs=3):
    """Saves audio as wave.

    NB: single channel only right now.

    Args:
      tag: str: label for this data
      audiodata: ndarray [Nsamples, Nframes, Nchannels]: audio data to
        be saved as wave. The data will be clipped to [-1.0, 1.0].
      step: int: training step
      sample_rate: sample rate of passed in audio buffer
      max_outputs: At most this many audio clips will be emitted at each
        step. Defaults to 3.
    """
    # tf.summary.audio expects the audio data to have floating values in
    # [-1.0, 1.0].
    audiodata = np.clip(np.array(audiodata), -1, 1)

    # Convert to tensor value as tf.summary.audio expects data to be a tensor.
    audio = tf.convert_to_tensor(audiodata, dtype=tf.float32)
    with self._as_default(self._event_writer):
      tf.summary.audio(
        name=tag,
        data=audio,
        sample_rate=sample_rate,
        step=step,
        max_outputs=max_outputs,
        encoding='wav',
      )

  def histogram(self, tag, values, step, bins=None):
    """Saves histogram of values.

    Args:
      tag: str: label for this data
      values: ndarray: will be flattened by this routine
      step: int: training step
      bins: number of bins in histogram
    """
    values = np.array(values)
    values = np.reshape(values, -1)
    with self._as_default(self._event_writer):
      tf.summary.histogram(name=tag, data=values, step=step, buckets=bins)

  def text(self, tag, textdata, step):
    """Saves a text summary.

    Args:
      tag: str: label for this data
      textdata: string
      step: int: training step
    Note: markdown formatting is rendered by tensorboard.
    """
    if not isinstance(textdata, (str, bytes)):
      raise ValueError('`textdata` should be of the type `str` or `bytes`.')
    with self._as_default(self._event_writer):
      tf.summary.text(name=tag, data=tf.constant(textdata), step=step)

  def write(self, tag, tensor, step, metadata=None):
    """Saves an arbitrary tensor summary.

    Useful when working with custom plugins or constructing a summary directly.

    Args:
      tag: str: label for this data
      tensor: ndarray: tensor data to save.
      step: int: training step
      metadata: Optional SummaryMetadata, as a proto or serialized bytes.
    Note: markdown formatting is rendered by tensorboard.
    """
    with self._as_default(self._event_writer):
      tf.summary.write(tag=tag, tensor=tensor, step=step, metadata=metadata)

  def hparams(self, hparams):
    """Saves hyper parameters.

    Args:
      hparams: Flat mapping from hyper parameter name to value.
    """

    with self._as_default(self._event_writer):
      hparams_api.hparams(hparams=_flatten_dict(hparams))



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for transformer modules."""

from absl.testing import absltest
from absl.testing import parameterized
from flax import nnx
import modules
import jax
import jax.numpy as jnp
import numpy as np


class EmbedderTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(
          vocab_size=10,
          embed_dim=4,
          inputs=[2, 3],
          expected=[[2.0, 2.0, 2.0, 2.0], [2.0, 2.0, 2.0, 2.0]],
      ),
  )
  def test_encode(self, vocab_size, embed_dim, inputs, expected):
    embedder = modules.Embedder(
        vocab_size=vocab_size,
        embed_dim=embed_dim,
        rngs=nnx.Rngs(params=0),
    )
    embedder.input_embedding.value = jnp.ones((vocab_size, embed_dim))
    output = embedder.encode(inputs)
    np.testing.assert_array_equal(output, jnp.array(expected))

  @parameterized.parameters(
      dict(
          vocab_size=5,
          embed_dim=2,
          inputs=[[1, 2]],
          expected=[[3.0, 3.0, 3.0, 3.0, 3.0]],
      ),
  )
  def test_decode(self, vocab_size, embed_dim, inputs, expected):
    embedder = modules.Embedder(
        vocab_size=vocab_size,
        embed_dim=embed_dim,
        rngs=nnx.Rngs(params=0),
    )
    embedder.input_embedding.value = jnp.ones((vocab_size, embed_dim))
    output = embedder.decode(jnp.array(inputs))
    np.testing.assert_array_equal(output, jnp.array(expected))


class AttentionTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(
          head_dim=2,
      ),
      dict(
          head_dim=20,
      ),
  )
  def test_head_dim(self, head_dim):
    attn = modules.Attention(
        num_heads=2,
        num_kv_heads=4,
        features=5,
        head_dim=head_dim,
        attn_type=modules.AttentionType.GLOBAL,
        rngs=nnx.Rngs(params=0),
    )

    self.assertEqual(attn.head_dim, head_dim)

  @parameterized.parameters(
      dict(
          num_heads=2,
          num_kv_heads=4,
          expected_use_qkv_einsum=False,
      ),
      dict(
          num_heads=3,
          num_kv_heads=3,
          expected_use_qkv_einsum=True,
      ),
  )
  def test_use_qkv_einsum(
      self,
      num_heads,
      num_kv_heads,
      expected_use_qkv_einsum,
  ):
    attn = modules.Attention(
        num_heads=num_heads,
        num_kv_heads=num_kv_heads,
        features=5,
        head_dim=8,
        attn_type=modules.AttentionType.GLOBAL,
        rngs=nnx.Rngs(params=0),
    )

    self.assertEqual(attn.use_qkv_einsum, expected_use_qkv_einsum)
    self.assertEqual(hasattr(attn, 'q_einsum'), not expected_use_qkv_einsum)
    self.assertEqual(hasattr(attn, 'kv_einsum'), not expected_use_qkv_einsum)

  @parameterized.parameters(
      dict(
          num_heads=2,
          head_dim=4,
          features=8,
          segment_pos=0,
          cache_size=3,
          batch_size=2,
          expected_cache_shape=(2, 3, 2, 4),
          expected_output_shape=(2, 1, 8),
      ),
  )
  def test_attention(
      self,
      num_heads,
      head_dim,
      features,
      segment_pos,
      cache_size,
      batch_size,
      expected_cache_shape,
      expected_output_shape,
  ):
    attn_mask = jnp.ones((batch_size, 1, cache_size))
    attn = modules.Attention(
        num_heads,
        num_heads,
        features,
        head_dim,
        modules.AttentionType.GLOBAL,
        rngs=nnx.Rngs(params=0),
    )
    cache = attn.init_cache(
        cache_size=cache_size,
        batch_size=batch_size,
        dtype=jnp.float32,
    )
    x = jnp.ones((batch_size, 1, features))
    cache, output = attn(x, jnp.array([[segment_pos]]), cache, attn_mask)

    self.assertEqual(cache['k'].shape, expected_cache_shape)
    self.assertEqual(output.shape, expected_output_shape)

  @parameterized.parameters(
      dict(
          sliding_window_size=2,
      ),
  )
  def test_sliding_window(self, sliding_window_size):
    num_heads = 2
    head_dim = 4
    features = 8
    segment_pos = 0
    cache_size = 3
    batch_size = 2
    attn_mask = jnp.ones((batch_size, 1, cache_size))
    x = jnp.ones((batch_size, 1, features))
    attn = modules.Attention(
        num_heads,
        num_heads,
        features,
        head_dim,
        modules.AttentionType.GLOBAL,
        rngs=nnx.Rngs(params=0),
    )
    cache = attn.init_cache(
        cache_size=cache_size,
        batch_size=batch_size,
        dtype=jnp.float32,
    )
    _, output = attn(x, jnp.array([[segment_pos]]), cache, attn_mask)
    sliding_attn = modules.Attention(
        num_heads,
        num_heads,
        features,
        head_dim,
        modules.AttentionType.LOCAL_SLIDING,
        sliding_window_size=sliding_window_size,
        rngs=nnx.Rngs(params=0),
    )
    _, sliding_output = sliding_attn(
        x, jnp.array([[segment_pos]]), cache, attn_mask
    )

    self.assertFalse((output == sliding_output).all())


class FeedForwardTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(
          features=2,
          hidden_dim=3,
          batch_size=2,
          expected_val=[11.72758674, 47.99916],
          expected_shape=(2, 1, 2),
      ),
  )
  def test_ffw(
      self, features, hidden_dim, batch_size, expected_val, expected_shape
  ):
    inputs = jnp.arange(1, batch_size+1)[:, None, None]
    inputs = jnp.repeat(inputs, features, axis=-1)
    ffw = modules.FeedForward(
        features=features,
        hidden_dim=hidden_dim,
        rngs=nnx.Rngs(params=0),
    )
    ffw.gating_einsum.value = jnp.ones((2, features, hidden_dim))
    ffw.linear.value = jnp.ones((hidden_dim, features))

    with jax.default_matmul_precision('float32'):
      outputs = ffw(inputs)

    np.testing.assert_array_almost_equal(outputs[:, 0, 0], expected_val)
    self.assertEqual(outputs.shape, expected_shape)


class BlockTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(
          num_heads=2,
          embed_dim=4,
          head_dim=6,
          cache_size=3,
          batch_size=2,
          use_post_attn_norm=False,
          use_post_ffw_norm=False,
          expected_cache_shape=(2, 3, 2, 6),
          expected_output_shape=(2, 1, 4),
      ),
  )
  def test_block(
      self,
      num_heads,
      embed_dim,
      head_dim,
      cache_size,
      batch_size,
      use_post_attn_norm,
      use_post_ffw_norm,
      expected_cache_shape,
      expected_output_shape,
  ):
    inputs = jnp.ones((batch_size, 1, embed_dim))
    attn_mask = jnp.ones((batch_size, 1, cache_size))

    block = modules.Block(
        num_heads,
        num_heads,
        embed_dim,
        head_dim,
        1,
        use_post_attn_norm,
        use_post_ffw_norm,
        modules.AttentionType.GLOBAL,
        rngs=nnx.Rngs(params=0),
    )
    cache = block.init_cache(
        cache_size=cache_size,
        batch_size=batch_size,
        dtype=jnp.float32,
    )

    new_cache, outputs = block(inputs, jnp.array([[0]]), cache, attn_mask)

    self.assertEqual(block.use_post_attn_norm, use_post_attn_norm)
    self.assertEqual(new_cache['k'].shape, expected_cache_shape)
    self.assertEqual(outputs.shape, expected_output_shape)

  @parameterized.parameters(
      dict(
          num_heads=1,
          embed_dim=1,
          head_dim=2,
          cache_size=1,
          batch_size=1,
      ),
  )
  def test_post_attention_norm(
      self,
      num_heads,
      embed_dim,
      head_dim,
      cache_size,
      batch_size,
  ):
    inputs = jnp.ones((batch_size, 1, embed_dim))
    attn_mask = jnp.ones((batch_size, 1, cache_size))

    normed_block = modules.Block(
        num_heads,
        num_heads,
        embed_dim,
        head_dim,
        1,
        True,
        False,  # use_post_ffw_norm
        modules.AttentionType.GLOBAL,
        rngs=nnx.Rngs(params=0),
    )
    unnormed_block = modules.Block(
        num_heads,
        num_heads,
        embed_dim,
        head_dim,
        1,
        False,
        False,  # use_post_ffw_norm
        modules.AttentionType.GLOBAL,
        rngs=nnx.Rngs(params=0),
    )

    # Ok to use the same cache for both blocks.
    cache = normed_block.init_cache(
        cache_size=cache_size,
        batch_size=batch_size,
        dtype=jnp.float32,
    )

    all_outputs = []
    for block in (normed_block, unnormed_block):
      _, outputs = block(inputs, jnp.array([[0]]), cache, attn_mask)
      all_outputs.append(outputs)

    normed_output, unnormed_output = all_outputs  # pylint: disable=unbalanced-tuple-unpacking
    self.assertFalse(jnp.not_equal(normed_output, unnormed_output).all())

  @parameterized.parameters(
      dict(
          num_heads=1,
          embed_dim=1,
          head_dim=2,
          cache_size=1,
          batch_size=1,
      ),
  )
  def test_post_ffw_norm(
      self,
      num_heads,
      embed_dim,
      head_dim,
      cache_size,
      batch_size,
  ):
    inputs = jnp.ones((batch_size, 1, embed_dim))
    attn_mask = jnp.ones((batch_size, 1, cache_size))

    normed_block = modules.Block(
        num_heads,
        num_heads,
        embed_dim,
        head_dim,
        1,
        True,
        True,  # use_post_ffw_norm
        modules.AttentionType.GLOBAL,
        rngs=nnx.Rngs(params=0),
    )
    unnormed_block = modules.Block(
        num_heads,
        num_heads,
        embed_dim,
        head_dim,
        1,
        False,
        False,  # use_post_ffw_norm
        modules.AttentionType.GLOBAL,
        rngs=nnx.Rngs(params=0),
    )

    # Ok to use the same cache for both blocks.
    cache = normed_block.init_cache(
        cache_size=cache_size,
        batch_size=batch_size,
        dtype=jnp.float32,
    )

    all_outputs = []
    for block in (normed_block, unnormed_block):
      _, outputs = block(inputs, jnp.array([[0]]), cache, attn_mask)
      all_outputs.append(outputs)

    normed_output, unnormed_output = all_outputs  # pylint: disable=unbalanced-tuple-unpacking
    self.assertFalse(jnp.not_equal(normed_output, unnormed_output).all())


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
import jax
import jax.numpy as jnp
import numpy as np

from temperature_sampler import temperature_sample


jax.config.update('jax_disable_most_optimizations', True)


class TestTemperatureSampler(absltest.TestCase):

  def test_temperature_sampler(self):
    tokens = jnp.array([[5, 0, 0, 0]], dtype=jnp.int32)
    cache = None
    key = jax.random.PRNGKey(0)

    def tokens_to_logits(tokens, cache):
      jax.debug.print('tokens: {}', tokens)
      logits = jax.nn.one_hot(tokens[..., -1:] + 1, 10)
      logits = jnp.where(logits < 0.5, float('-inf'), logits)
      logits = logits.squeeze(axis=1)
      return logits, cache

    new_tokens = temperature_sample(
        tokens, cache, tokens_to_logits, key, topk=5
    )

    np.testing.assert_array_equal(new_tokens, [[5, 6, 7, 8]])


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.linen.partitioning."""

import jax
import jax.numpy as jnp
from absl.testing import absltest, parameterized
from jax import random, sharding
from jax.experimental import mesh_utils

from flax import linen as nn
from flax.core import freeze, unfreeze
from flax.linen import partitioning

mock = absltest.mock

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()

# Testing constants.
AXIS_RULES_1 = (('foo', 'data'), ('bar', 'model'), ('baz', None))
AXIS_RULES_2 = (('foo', 'model'), ('bar', None), ('baz', 'data'))


class PartitioningTest(parameterized.TestCase):
  def test_axis_rules(self):
    self.assertEqual(partitioning._axis_rules.rules, ())
    partitioning.set_axis_rules(AXIS_RULES_1)
    self.assertEqual(partitioning._axis_rules.rules, AXIS_RULES_1)
    self.assertEqual(partitioning.get_axis_rules(), AXIS_RULES_1)
    partitioning.set_axis_rules(())

  def test_axis_rules_context(self):
    partitioning.set_axis_rules(AXIS_RULES_1)
    self.assertEqual(partitioning.get_axis_rules(), AXIS_RULES_1)
    with partitioning.axis_rules(AXIS_RULES_2):
      self.assertEqual(partitioning.get_axis_rules(), AXIS_RULES_2)
    self.assertEqual(partitioning.get_axis_rules(), AXIS_RULES_1)

  def test_logical_to_mesh_axes(self):
    axes_0 = ('foo', 'bar')
    # direct rule assignment
    self.assertEqual(
      partitioning.logical_to_mesh_axes(axes_0, rules=AXIS_RULES_1),
      ('data', 'model'),
    )
    # axis rules context
    with partitioning.axis_rules(AXIS_RULES_1):
      self.assertEqual(
        partitioning.logical_to_mesh_axes(axes_0), ('data', 'model')
      )
      # nested context
      with partitioning.axis_rules(AXIS_RULES_2):
        self.assertEqual(
          partitioning.logical_to_mesh_axes(axes_0), ('model', None)
        )
    # duplicated logical names
    with partitioning.axis_rules(AXIS_RULES_1):
      with self.assertRaises(ValueError):
        partitioning.logical_to_mesh_axes(('foo', 'foo', 'baz'))

  def test_logical_to_mesh_axes_priorities(self):
    p_rules = (('foo', 'model'), ('bar', 'model'), ('baz', 'data'))
    with partitioning.axis_rules(p_rules):
      self.assertEqual(
        partitioning.logical_to_mesh_axes(('foo', 'bar', 'baz')),
        ('model', None, 'data'),
      )
      self.assertEqual(
        partitioning.logical_to_mesh_axes(('bar', 'foo', 'baz')),
        (None, 'model', 'data'),
      )
      self.assertEqual(
        partitioning.logical_to_mesh_axes(('baz', 'bar', 'foo')),
        ('data', None, 'model'),
      )
      self.assertEqual(
        partitioning.logical_to_mesh_axes(('baz', 'bar', 'foo', 'unassigned')),
        ('data', None, 'model', None),
      )

  @parameterized.parameters(
      dict(
          rules=(('a', ('model', 'data')), ('b', 'data')),
          axes=('a', 'b'),
          expected=(('model', 'data'), None),
      ),
      dict(
          rules=(('a', ('model', 'replica')), ('b', 'data')),
          axes=('a', 'b'),
          expected=(('model', 'replica'), 'data'),
      ),
      dict(
          rules=(('a', ('model', 'replica')), ('b', ('data', 'model'))),
          axes=('a', 'b'),
          expected=(('model', 'replica'), None),
      ),
      dict(
          rules=(('a', ('model', 'replica')), ('b', 'model')),
          axes=('a', 'b', 'c'),
          expected=(('model', 'replica'), None, None),
      ),
      dict(rules=(), axes=('a', 'b', 'c'), expected=(None, None, None)),
      dict(
          rules=(('a', None), ('a', 'model')),
          axes=('a', 'b'),
          expected=(None, None),
      ),
      dict(
          rules=(
              ('baz', 'data'),
              ('bar', None),
              ('foo', 'model'),
              ('foo', 'data'),
          ),
          axes=('baz', 'bar', 'foo'),
          expected=('data', None, 'model'),
      ),
      dict(
          rules=(('baz', 'data'), ('foo', ('model', 'emb'))),
          axes=('baz', jax.sharding.PartitionSpec.UNCONSTRAINED, 'foo'),
          expected=(
              'data',
              jax.sharding.PartitionSpec.UNCONSTRAINED,
              ('model', 'emb'),
          ),
      ),
  )
  def test_logical_to_mesh_axes_cases(self, rules, axes, expected):
    with partitioning.axis_rules(rules):
      result = partitioning.logical_to_mesh_axes(axes)
    self.assertEqual(result, expected)

  @mock.patch('flax.linen.spmd._with_sharding_constraint')
  def test_with_sharding_constraint(self, wsc_fn):
    unconstrained = jax.sharding.PartitionSpec.UNCONSTRAINED
    arr = jnp.ones((2, 2))
    axes = ('foo', 'bar')
    partitioning.set_axis_rules(())
    _ = partitioning.with_sharding_constraint(arr, axes)
    wsc_fn.assert_not_called()
    with partitioning.axis_rules(AXIS_RULES_1):
      _ = partitioning.with_sharding_constraint(arr, None)
      wsc_fn.assert_not_called()
      _ = partitioning.with_sharding_constraint(arr, axes)
      wsc_fn.assert_called_with(
          arr, jax.sharding.PartitionSpec('data', 'model'), mesh=None
      )
      _ = partitioning.with_sharding_constraint(arr, ('foo', unconstrained))
      wsc_fn.assert_called_with(
          arr, jax.sharding.PartitionSpec('data', unconstrained), mesh=None
      )

  @mock.patch('flax.linen.spmd._with_sharding_constraint')
  def test_with_sharding_constraint_fallback(self, wsc_fn):
    arr = jnp.ones((2, 2))
    with partitioning.axis_rules(AXIS_RULES_1):
      _ = partitioning.with_sharding_constraint(arr, ('foo', 'not_recognized'))
      wsc_fn.assert_called_with(
        arr, jax.sharding.PartitionSpec('data', None), mesh=None
      )
      wsc_fn.reset_mock()
      _ = partitioning.with_sharding_constraint(
        arr,
        ('foo', 'not_recognized'),
        fallback=partitioning.RulesFallback.AXIS_IS_UNSHARDED,
      )
      wsc_fn.assert_called_with(
        arr, jax.sharding.PartitionSpec('data', None), mesh=None
      )
      wsc_fn.reset_mock()
      with self.assertRaises(ValueError):
        _ = partitioning.with_sharding_constraint(
          arr,
          ('foo', 'not_recognized'),
          fallback=partitioning.RulesFallback.RAISE_ERROR,
        )
      wsc_fn.assert_not_called()
      _ = partitioning.with_sharding_constraint(
        arr,
        ('foo', 'not_recognized'),
        fallback=partitioning.RulesFallback.NO_CONSTRAINT,
      )
      wsc_fn.assert_not_called()

  @parameterized.parameters(dict(axes_spec=None), dict(axes_spec=()))
  def test_param_with_axes_no_axes(self, axes_spec):
    class ParamTest(nn.Module):
      @nn.compact
      def __call__(self, x):
        foo = partitioning.param_with_axes(
          'foo',
          lambda k, s, d: jnp.zeros(s, d),
          (2, 2),
          x.dtype,
          axes=axes_spec,
        )
        return x + foo

    k = random.key(0)
    x = jnp.ones((2, 2))
    _ = ParamTest().init(k, x)

  def test_param_with_axes(self):
    class ParamTest(nn.Module):
      @nn.compact
      def __call__(self, x):
        foo = partitioning.param_with_axes(
          'foo',
          lambda k, s, d: jnp.zeros(s, d),
          (2, 2),
          x.dtype,
          axes=('foo', 'bar'),
        )
        return x + foo

    p_rules = (('foo', 'model'), ('bar', 'data'), ('baz', None))
    k = random.key(0)
    x = jnp.ones((2, 2))
    with partitioning.axis_rules(p_rules):
      variables = ParamTest().init(k, x)
    self.assertIn('params', variables)
    self.assertIn('params_axes', variables)
    self.assertEqual(
      variables['params_axes']['foo_axes'],
      partitioning.AxisMetadata(names=('foo', 'bar')),
    )
    logical_axis_names = partitioning.get_axis_names(variables['params_axes'])
    self.assertEqual(
      logical_axis_names, {'foo': jax.sharding.PartitionSpec('foo', 'bar')}
    )

  def test_param_pytree_with_axes(self):
    def init_fn(k, s, d):
      del k
      return {'a': jnp.zeros(s, d), 'b': (jnp.zeros(s, d), jnp.zeros(s, d))}

    axes = {'a': ('foo', 'bar'), 'b': (('foo', 'bar'), ('bar', 'foo'))}

    class ParamTest(nn.Module):
      @nn.compact
      def __call__(self, x):
        foo = partitioning.param_with_axes(
          'foo', init_fn, (2, 2), x.dtype, axes=axes
        )
        return x + foo['a']

    p_rules = (('foo', 'model'), ('bar', 'data'), ('baz', None))
    k = random.key(0)
    x = jnp.ones((2, 2))
    with partitioning.axis_rules(p_rules):
      variables = ParamTest().init(k, x)
    self.assertIn('params', variables)
    self.assertIn('params_axes', variables)
    self.assertEqual(
      variables['params_axes']['foo_axes'],
      partitioning.AxisMetadata(names=axes),
    )
    logical_axis_names = partitioning.get_axis_names(variables['params_axes'])
    expected = freeze(
      {
        'foo': {
          'a': jax.sharding.PartitionSpec('foo', 'bar'),
          'b': (
            jax.sharding.PartitionSpec('foo', 'bar'),
            jax.sharding.PartitionSpec('bar', 'foo'),
          ),
        }
      }
    )
    self.assertEqual(logical_axis_names, expected)

  @parameterized.parameters(dict(axes_spec=None), dict(axes_spec=()))
  def test_variable_with_axes_no_axes(self, axes_spec):
    class VarTest(nn.Module):
      @nn.compact
      def __call__(self, x):
        foo = partitioning.variable_with_axes(
          'test', 'foo', jnp.zeros, (2, 2), x.dtype, axes=axes_spec
        )
        return x + foo.value

    k = random.key(0)
    x = jnp.ones((2, 2))
    _ = VarTest().init(k, x)

  def test_variable_with_empty_tuple_has_empty_axes(self):
    class VarTest(nn.Module):
      @nn.compact
      def __call__(self, x):
        foo = partitioning.variable_with_axes(
          'test', 'foo', jnp.zeros, (2, 2), x.dtype, axes=()
        )
        return x + foo.value

    k = random.key(0)
    x = jnp.ones((2, 2))
    variables = VarTest().init(k, x)
    logical_axis_names = partitioning.get_axis_names(variables['test_axes'])
    self.assertEqual(logical_axis_names, {'foo': jax.sharding.PartitionSpec()})

  def test_variable_with_axes(self):
    class VarTest(nn.Module):
      @nn.compact
      def __call__(self, x):
        foo = partitioning.variable_with_axes(
          'test', 'foo', jnp.zeros, (2, 2), x.dtype, axes=('foo', 'bar')
        )
        return x + foo.value

    p_rules = (('foo', 'model'), ('bar', 'data'), ('baz', None))
    k = random.key(0)
    x = jnp.ones((2, 2))
    with partitioning.axis_rules(p_rules):
      variables = VarTest().init(k, x)
    self.assertIn('test', variables)
    self.assertIn('test_axes', variables)
    self.assertEqual(
      variables['test_axes']['foo_axes'],
      partitioning.AxisMetadata(names=('foo', 'bar')),
    )
    logical_axis_names = partitioning.get_axis_names(variables['test_axes'])
    self.assertEqual(
      logical_axis_names, {'foo': jax.sharding.PartitionSpec('foo', 'bar')}
    )

  @mock.patch('flax.linen.partitioning._with_sharding_constraint')
  def test_variable_with_axes_fallback(self, wsc_fn):
    class VarTest(nn.Module):
      @nn.compact
      def __call__(self, x):
        foo = partitioning.variable_with_axes(
          'test',
          'foo',
          jnp.zeros,
          (2, 2),
          x.dtype,
          axes=('foo', 'bar'),
          fallback=partitioning.RulesFallback.NO_CONSTRAINT,
        )
        return x + foo.value

    p_rules = (
      # No rule for 'foo':
      ('bar', 'data'),
      ('baz', None),
    )
    k = random.key(0)
    x = jnp.ones((2, 2))
    with partitioning.axis_rules(p_rules):
      variables = VarTest().init(k, x)

    wsc_fn.assert_not_called()
    self.assertIn('test', variables)
    self.assertIn('test_axes', variables)
    self.assertEqual(
      variables['test_axes']['foo_axes'],
      partitioning.AxisMetadata(names=('foo', 'bar')),
    )
    logical_axis_names = partitioning.get_axis_names(variables['test_axes'])
    self.assertEqual(
      logical_axis_names, {'foo': jax.sharding.PartitionSpec('foo', 'bar')}
    )

  def test_scan_with_axes(self):
    # MLP Hparams
    B, L, E = 8, 4, 32  # pylint: disable=invalid-name
    # fake inputs
    x = jnp.ones((B, E))
    k = random.key(0)

    class SinDot(nn.Module):
      depth: int

      @nn.compact
      def __call__(self, x):
        W1 = partitioning.param_with_axes(  # pylint: disable=invalid-name
          'W1',
          nn.initializers.xavier_normal(),
          (x.shape[-1], self.depth),
          axes=('emb', 'mlp'),
        )
        W2 = partitioning.param_with_axes(  # pylint: disable=invalid-name
          'W2',
          nn.initializers.xavier_normal(),
          (self.depth, x.shape[-1]),
          axes=('mlp', 'emb'),
        )
        y = jnp.dot(jnp.sin(jnp.dot(x, W1)), W2)
        _ = partitioning.variable_with_axes(
          'stats', 'y_st', lambda: y, axes=('batch', 'emb')
        )
        # scan expects a (carry, out) return signature.
        return y, None

    class Scanned(nn.Module):
      num_layers: int
      depth: int

      @nn.compact
      def __call__(self, x):
        scanned_sindot = partitioning.scan_with_axes(
          SinDot,
          in_axes=(),
          variable_axes={'params': 0, 'stats': 1},
          split_rngs={'params': True},
          axis_name='layer',
          axes_collections=('params', 'stats'),
          length=self.num_layers,
        )(self.depth, name='scanned_layer')
        y, _ = scanned_sindot(x)
        # test calling again to test metadata compatibility across calls
        _, _ = scanned_sindot(x)
        return y

    p_rules = (('emb', 'data'), ('mlp', 'model'), ('batch', 'data'))
    with partitioning.axis_rules(p_rules):
      variables = Scanned(L, E).init(k, x)

      # Ensure that the module can be called when 'params_axes' is not mutable.
      Scanned(L, E).apply(variables, x)
    self.assertIn('params', variables)
    self.assertIn('params_axes', variables)
    self.assertIn('stats', variables)
    self.assertIn('stats_axes', variables)
    self.assertEqual(
      variables['params_axes']['scanned_layer']['W1_axes'],
      partitioning.AxisMetadata(names=('layer', 'emb', 'mlp')),
    )
    logical_axis_names = partitioning.get_axis_names(variables['params_axes'])
    self.assertEqual(
      logical_axis_names,
      {
        'scanned_layer': {
          'W1': jax.sharding.PartitionSpec('layer', 'emb', 'mlp'),
          'W2': jax.sharding.PartitionSpec('layer', 'mlp', 'emb'),
        }
      },
    )
    logical_axis_names = partitioning.get_axis_names(variables['stats_axes'])
    self.assertEqual(
      logical_axis_names,
      {
        'scanned_layer': {
          'y_st': jax.sharding.PartitionSpec('batch', 'layer', 'emb')
        }
      },
    )

  def test_vmap_with_axes(self):
    class Foo(nn.Module):
      @nn.compact
      def __call__(self, x):
        return (
          partitioning.param_with_axes(
            'w', jax.nn.initializers.uniform(), [4, 3], axes=('out', 'in')
          )
          @ x
        )

    class Vmapped(nn.Module):
      @nn.compact
      def __call__(self, x):
        FooVmapped = partitioning.vmap_with_axes(  # pylint: disable=invalid-name
          Foo,
          variable_axes={
            'params': 1,
          },
          split_rngs={'params': True},
          partitioning_axis_names={'params': 'vmap_axis'},
        )
        return FooVmapped(name='foo_vmapped')(x)

    p_rules = (('out', None), ('in', 'data'), ('vmap_axis', 'model'))

    # check that regular Food module is correct
    with partitioning.axis_rules(p_rules):
      variables = Foo().init(jax.random.key(0), jnp.array([1, 2, 3]))
    variables = unfreeze(variables)
    variables['params'] = jax.tree_util.tree_map(
      lambda x: x.shape, variables['params']
    )
    self.assertDictEqual(
      variables,
      {
        'params': {'w': (4, 3)},
        'params_axes': {
          'w_axes': partitioning.AxisMetadata(names=('out', 'in'))
        },
      },
    )

    # check that FooVmapped adds 'vmap_axis' to axis 1
    with partitioning.axis_rules(p_rules):
      variables = Vmapped().init(
        jax.random.key(0), jnp.array([[1, 2, 3], [4, 5, 6]])
      )
    variables = unfreeze(variables)
    variables['params'] = jax.tree_util.tree_map(
      lambda x: x.shape, variables['params']
    )
    self.assertDictEqual(
      variables,
      {
        'params': {'foo_vmapped': {'w': (4, 2, 3)}},
        'params_axes': {
          'foo_vmapped': {
            'w_axes': partitioning.AxisMetadata(
              names=('out', 'vmap_axis', 'in')
            )
          }
        },
      },
    )

  def test_logical_with_mesh_and_rules(self):
    devices = mesh_utils.create_device_mesh((jax.local_device_count(), 1))
    mesh = sharding.Mesh(devices, ('in', 'out'))
    test = self
    rules = (('a', 'in'), ('b', 'out'))

    class Foo(nn.Module):
      @nn.compact
      def __call__(self, x):
        kernel_init = nn.with_logical_partitioning(
          nn.initializers.ones_init(), ('a', 'b'), mesh=mesh, rules=rules
        )
        kernel = self.param('kernel', kernel_init, (x.shape[-1], 2))
        kernel_box = self.get_variable('params', 'kernel')
        test.assertIsInstance(kernel_box, nn.Partitioned)
        test.assertEqual(kernel_box.names, ('a', 'b'))
        return x @ kernel

    @jax.jit
    def create_state():
      module = Foo()
      variables = module.init(random.key(0), jnp.zeros((8, 4)))
      logical_spec = nn.get_partition_spec(variables)
      shardings = nn.logical_to_mesh_sharding(logical_spec, mesh, rules)
      variables = jax.lax.with_sharding_constraint(variables, shardings)
      return variables

    variables = create_state()
    self.assertEqual(variables['params']['kernel'].names, ('a', 'b'))
    self.assertIs(variables['params']['kernel'].mesh, mesh)
    self.assertEqual(variables['params']['kernel'].rules, rules)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing as tp

import jax
import jax.numpy as jnp
from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from flax import linen
from flax import nnx
from flax.typing import Dtype


class TestLinenConsistency(parameterized.TestCase):
  @parameterized.product(
    dtype=[jnp.float32, jnp.float16],
    param_dtype=[jnp.float32, jnp.float16],
    use_fast_variance=[True, False],
    mask=[None, np.array([True, False, True, False, True])],
  )
  def test_nnx_linen_batchnorm_equivalence(
    self,
    dtype: tp.Optional[Dtype],
    param_dtype: Dtype,
    use_fast_variance: bool,
    mask: tp.Optional[np.ndarray],
  ):
    class NNXModel(nnx.Module):
      def __init__(self, dtype, param_dtype, use_fast_variance, rngs):
        self.norm_layer = nnx.BatchNorm(
          5,
          use_running_average=False,
          dtype=dtype,
          param_dtype=param_dtype,
          use_fast_variance=use_fast_variance,
          rngs=rngs,
        )
        self.linear = nnx.Linear(
          5, 4, dtype=dtype, param_dtype=param_dtype, rngs=rngs
        )

      def __call__(self, x, *, mask=None):
        x = self.norm_layer(x, mask=mask)
        x = self.linear(x)
        return x

    class LinenModel(linen.Module):
      dtype: tp.Optional[Dtype] = None
      param_dtype: Dtype = jnp.float32
      use_fast_variance: bool = True

      def setup(self):
        self.norm_layer = linen.BatchNorm(
          use_running_average=False,
          dtype=self.dtype,
          param_dtype=self.param_dtype,
          use_fast_variance=use_fast_variance,
        )
        self.linear = linen.Dense(
          4, dtype=self.dtype, param_dtype=self.param_dtype
        )

      def __call__(self, x, *, mask=None):
        x = self.norm_layer(x, mask=mask)
        x = self.linear(x)
        return x

    rngs = nnx.Rngs(42)
    x = jax.random.normal(jax.random.key(0), (10, 5))

    linen_model = LinenModel(
      dtype=dtype, param_dtype=param_dtype, use_fast_variance=use_fast_variance
    )
    variables = linen_model.init(jax.random.key(1), x)
    linen_out, batch_stats = linen_model.apply(
      variables, x, mask=mask, mutable=['batch_stats']
    )

    nnx_model = NNXModel(
      dtype=dtype,
      param_dtype=param_dtype,
      use_fast_variance=use_fast_variance,
      rngs=rngs,
    )
    nnx_model.linear.kernel.value = variables['params']['linear']['kernel']
    nnx_model.linear.bias.value = variables['params']['linear']['bias']

    nnx_out = nnx_model(x, mask=mask)
    np.testing.assert_array_equal(linen_out, nnx_out)

  @parameterized.product(
    dtype=[jnp.float32, jnp.float16],
    param_dtype=[jnp.float32, jnp.float16],
    use_fast_variance=[True, False],
    mask=[None, np.array([True, False, True, False, True])],
  )
  def test_nnx_linen_layernorm_equivalence(
    self,
    dtype: tp.Optional[Dtype],
    param_dtype: Dtype,
    use_fast_variance: bool,
    mask: tp.Optional[np.ndarray],
  ):
    class NNXModel(nnx.Module):
      def __init__(self, dtype, param_dtype, use_fast_variance, rngs):
        self.norm_layer = nnx.LayerNorm(
          5,
          dtype=dtype,
          param_dtype=param_dtype,
          use_fast_variance=use_fast_variance,
          rngs=rngs,
        )
        self.linear = nnx.Linear(
          5, 4, dtype=dtype, param_dtype=param_dtype, rngs=rngs
        )

      def __call__(self, x, *, mask=None):
        x = self.norm_layer(x, mask=mask)
        x = self.linear(x)
        return x

    class LinenModel(linen.Module):
      dtype: tp.Optional[Dtype] = None
      param_dtype: Dtype = jnp.float32
      use_fast_variance: bool = True

      def setup(self):
        self.norm_layer = linen.LayerNorm(
          dtype=self.dtype,
          param_dtype=self.param_dtype,
          use_fast_variance=self.use_fast_variance,
        )
        self.linear = linen.Dense(
          4, dtype=self.dtype, param_dtype=self.param_dtype
        )

      def __call__(self, x, *, mask=None):
        x = self.norm_layer(x, mask=mask)
        x = self.linear(x)
        return x

    rngs = nnx.Rngs(42)
    x = jax.random.normal(jax.random.key(0), (10, 5))

    linen_model = LinenModel(
      dtype=dtype, param_dtype=param_dtype, use_fast_variance=use_fast_variance
    )
    variables = linen_model.init(jax.random.key(1), x)
    linen_out = linen_model.apply(variables, x, mask=mask)

    nnx_model = NNXModel(
      dtype=dtype,
      param_dtype=param_dtype,
      use_fast_variance=use_fast_variance,
      rngs=rngs,
    )
    nnx_model.linear.kernel.value = variables['params']['linear']['kernel']
    nnx_model.linear.bias.value = variables['params']['linear']['bias']

    nnx_out = nnx_model(x, mask=mask)
    np.testing.assert_array_equal(linen_out, nnx_out)

  @parameterized.product(
    dtype=[jnp.float32, jnp.float16],
    param_dtype=[jnp.float32, jnp.float16],
    use_fast_variance=[True, False],
    mask=[None, np.array([True, False, True, False, True])],
  )
  def test_nnx_linen_rmsnorm_equivalence(
    self,
    dtype: tp.Optional[Dtype],
    param_dtype: Dtype,
    use_fast_variance: bool,
    mask: tp.Optional[np.ndarray],
  ):
    class NNXModel(nnx.Module):
      def __init__(self, dtype, param_dtype, use_fast_variance, rngs):
        self.norm_layer = nnx.RMSNorm(
          5,
          dtype=dtype,
          param_dtype=param_dtype,
          use_fast_variance=use_fast_variance,
          rngs=rngs,
        )
        self.linear = nnx.Linear(
          5, 4, dtype=dtype, param_dtype=param_dtype, rngs=rngs
        )

      def __call__(self, x, *, mask=None):
        x = self.norm_layer(x, mask=mask)
        x = self.linear(x)
        return x

    class LinenModel(linen.Module):
      dtype: tp.Optional[Dtype] = None
      param_dtype: Dtype = jnp.float32
      use_fast_variance: bool = True

      def setup(self):
        self.norm_layer = linen.RMSNorm(
          dtype=self.dtype,
          param_dtype=self.param_dtype,
          use_fast_variance=self.use_fast_variance,
        )
        self.linear = linen.Dense(
          4, dtype=self.dtype, param_dtype=self.param_dtype
        )

      def __call__(self, x, *, mask=None):
        x = self.norm_layer(x, mask=mask)
        x = self.linear(x)
        return x

    rngs = nnx.Rngs(42)
    x = jax.random.normal(jax.random.key(0), (10, 5))

    linen_model = LinenModel(
      dtype=dtype, param_dtype=param_dtype, use_fast_variance=use_fast_variance
    )
    variables = linen_model.init(jax.random.key(1), x)
    linen_out = linen_model.apply(variables, x, mask=mask)

    nnx_model = NNXModel(
      dtype=dtype,
      param_dtype=param_dtype,
      use_fast_variance=use_fast_variance,
      rngs=rngs,
    )
    nnx_model.linear.kernel.value = variables['params']['linear']['kernel']
    nnx_model.linear.bias.value = variables['params']['linear']['bias']

    nnx_out = nnx_model(x, mask=mask)
    np.testing.assert_array_equal(linen_out, nnx_out)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.jax_utils."""

from functools import partial

from absl.testing import absltest
import chex
import jax
import numpy as np
from absl.testing import parameterized

from flax import jax_utils

NDEV = 4


def setUpModule():
  chex.set_n_cpu_devices(NDEV)


class PadShardUnpadTest(chex.TestCase):
  BATCH_SIZES = [NDEV, NDEV + 1, NDEV - 1, 5 * NDEV, 5 * NDEV + 1, 5 * NDEV - 1]
  DTYPES = [np.float32, np.uint8, jax.numpy.bfloat16, np.int32]

  def tearDown(self):
    chex.clear_trace_counter()
    super().tearDown()

  @parameterized.product(dtype=DTYPES, bs=BATCH_SIZES)
  def test_basics(self, dtype, bs):
    # Just tests that basic calling works without exploring caveats.
    @partial(jax_utils.pad_shard_unpad, static_argnums=())
    def add(a, b):
      return a + b

    x = np.arange(bs, dtype=dtype)
    y = add(x, 10 * x)
    chex.assert_type(y.dtype, x.dtype)
    np.testing.assert_allclose(np.float64(y), np.float64(x + 10 * x))

  @parameterized.product(dtype=DTYPES, bs=BATCH_SIZES)
  def test_trees(self, dtype, bs):
    # Just tests that basic calling works without exploring caveats.
    @partial(jax_utils.pad_shard_unpad, static_argnums=())
    def add(a, b):
      return a['a'] + b[0]

    x = np.arange(bs, dtype=dtype)
    y = add(dict(a=x), (10 * x,))
    chex.assert_type(y.dtype, x.dtype)
    np.testing.assert_allclose(np.float64(y), np.float64(x + 10 * x))

  @parameterized.parameters(DTYPES)
  def test_min_device_batch_avoids_recompile(self, dtype):
    @partial(jax_utils.pad_shard_unpad, static_argnums=())
    @jax.jit
    @chex.assert_max_traces(n=1)
    def add(a, b):
      return a + b

    chex.clear_trace_counter()

    for bs in self.BATCH_SIZES:
      x = np.arange(bs, dtype=dtype)
      y = add(x, 10 * x, min_device_batch=9)  # pylint: disable=unexpected-keyword-arg
      chex.assert_type(y.dtype, x.dtype)
      np.testing.assert_allclose(np.float64(y), np.float64(x + 10 * x))

  @parameterized.product(dtype=DTYPES, bs=BATCH_SIZES)
  def test_static_argnum(self, dtype, bs):
    @partial(jax_utils.pad_shard_unpad, static_argnums=(1,))
    def add(a, b):
      return a + b

    x = np.arange(bs, dtype=dtype)
    y = add(x, 10)
    chex.assert_type(y.dtype, x.dtype)
    np.testing.assert_allclose(np.float64(y), np.float64(x + 10))

  @parameterized.product(dtype=DTYPES, bs=BATCH_SIZES)
  def test_static_argnames(self, dtype, bs):
    # In this test, leave static_argnums at the default value too, in order to
    # test the default/most canonical path where `params` are the first arg.
    @partial(jax_utils.pad_shard_unpad, static_argnames=('b',))
    def add(params, a, *, b):
      return params * a + b

    x = np.arange(bs, dtype=dtype)
    y = add(5, x, b=10)
    chex.assert_type(y.dtype, x.dtype)
    np.testing.assert_allclose(np.float64(y), np.float64(5 * x + 10))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing as tp

from absl.testing import absltest
import jax
import jax.numpy as jnp
import numpy as np

from flax import nnx

A = tp.TypeVar('A')


class TestIntegration(absltest.TestCase):
  def test_shared_modules(self):
    class Block(nnx.Module):
      def __init__(self, linear: nnx.Linear, *, rngs):
        self.linear = linear
        self.bn = nnx.BatchNorm(2, rngs=rngs)

      def __call__(self, x):
        x = self.linear(x)
        x = self.bn(x)
        return nnx.relu(x)

    class Model(nnx.Module):
      def __init__(self, *, rngs):
        shared = nnx.Linear(2, 2, rngs=rngs)
        self.block1 = Block(shared, rngs=rngs)
        self.block2 = Block(shared, rngs=rngs)

      def __call__(self, x):
        x = self.block1(x)
        x = self.block2(x)
        return x

    @nnx.jit
    def train_step(model: Model, x, y):
      @nnx.grad
      def loss_fn(model: Model):
        y_pred = model(x)
        return jnp.mean((y - y_pred) ** 2)

      grads = loss_fn(model)
      nnx.update(
        model,
        jax.tree_util.tree_map(
          lambda w, g: w - 0.1 * g, nnx.state(model, nnx.Param), grads
        ),
      )

    model = Model(rngs=nnx.Rngs(0))

    x = np.random.uniform(size=(4, 2))
    y = np.random.uniform(size=(4, 2))
    model.set_attributes(use_running_average=False)

    for _i in range(3):
      train_step(model, x, y)

    assert model.block1.linear is model.block2.linear
    assert model.block1.linear.bias is not None
    assert model.block1.bn is not model.block2.bn

  def test_shared_modules_pure(self):
    class Block(nnx.Module):
      def __init__(self, linear: nnx.Linear, *, rngs: nnx.Rngs):
        self.linear = linear
        self.bn = nnx.BatchNorm(2, rngs=rngs)

      def __call__(self, x):
        x = self.linear(x)
        x = self.bn(x)
        return nnx.relu(x)

    class Model(nnx.Module):
      def __init__(self, *, rngs: nnx.Rngs):
        shared = nnx.Linear(2, 2, rngs=rngs)
        self.block1 = Block(shared, rngs=rngs)
        self.block2 = Block(shared, rngs=rngs)

      def __call__(self, x):
        x = self.block1(x)
        x = self.block2(x)
        return x

    @jax.jit
    def train_step(state: nnx.State, graphdef: nnx.GraphDef[Model], x, y):
      model = nnx.merge(graphdef, state)
      model.set_attributes(use_running_average=False)

      @nnx.grad
      def loss_fn(model: Model):
        y_pred = model(x)
        return jnp.mean((y - y_pred) ** 2)

      grads = loss_fn(model)
      nnx.update(
        model,
        jax.tree_util.tree_map(
          lambda w, g: w - 0.1 * g, nnx.state(model, nnx.Param), grads
        ),
      )

      return nnx.split(model)

    graphdef: nnx.GraphDef[Model]
    graphdef, state = nnx.split(Model(rngs=nnx.Rngs(0)))

    x = np.random.uniform(size=(4, 2))
    y = np.random.uniform(size=(4, 2))

    for _i in range(3):
      graphdef, state = train_step(state, graphdef, x, y)

    model = nnx.merge(graphdef, state)

    assert model.block1.linear.bias is not None
    assert model.block2.linear.bias is not None
    assert model.block1.linear.kernel is model.block2.linear.kernel
    assert model.block1.linear.bias is model.block2.linear.bias
    assert model.block1.bn is not model.block2.bn

  def test_stateful_example(self):
    class State(nnx.Variable[A]):
      pass

    class Linear(nnx.Module):
      def __init__(self, din: int, dout: int, *, rngs: nnx.Rngs):
        key = rngs.params()
        self.w = nnx.Param(jax.random.uniform(key, (din, dout)))
        self.b = nnx.Param(jnp.zeros((dout,)))
        self.count = State(0)

      def __call__(self, x):
        self.count.value += 1
        return x @ self.w.value + self.b.value[None]

    model = Linear(din=12, dout=2, rngs=nnx.Rngs(0))
    # forward pass
    x = jnp.ones((8, 12))
    y = model(x)
    assert model.count.value == 1

    @nnx.jit
    def train_step(model, x, y):
      def loss_fn(model):
        y_pred = model(x)
        return jax.numpy.mean((y_pred - y) ** 2)

      # compute gradient
      grads: nnx.State = nnx.grad(loss_fn, wrt=nnx.Param)(model)
      # SGD update
      nnx.update(
        model,
        jax.tree_util.tree_map(
          lambda w, g: w - 0.1 * g, nnx.state(model, nnx.Param), grads
        ),
      )

    # execute the training step
    train_step(model, x, y)
    assert model.count.value == 2

  def test_functional_example(self):
    class Count(nnx.Variable[A]):
      pass

    class Linear(nnx.Module):
      def __init__(self, din: int, dout: int, *, rngs: nnx.Rngs):
        key = rngs.params()
        self.w = nnx.Param(jax.random.uniform(key, (din, dout)))
        self.b = nnx.Param(jnp.zeros((dout,)))
        self.count = Count(0)

      def __call__(self, x):
        self.count.value += 1
        return x @ self.w.value + self.b.value[None]

    model = Linear(din=12, dout=2, rngs=nnx.Rngs(0))
    # forward pass
    x = jnp.ones((8, 12))
    y = model(x)
    assert model.count.value == 1

    graphdef, params, counts = nnx.split(model, nnx.Param, Count)

    @jax.jit
    def train_step(params, counts, x, y):
      def loss_fn(params):
        y_pred, (_, updates) = graphdef.apply(params, counts)(x)
        loss = jax.numpy.mean((y_pred - y) ** 2)
        return loss, updates.filter(Count)

      # compute gradient
      grads, counts = jax.grad(loss_fn, has_aux=True)(params)
      # SGD update
      params = jax.tree_util.tree_map(lambda w, g: w - 0.1 * g, params, grads)

      return params, counts

    # execute the training step
    params, counts = train_step(params, counts, x, y)
    model = nnx.merge(graphdef, params, counts)
    assert model.count.value == 2

  def test_intermediates_example(self):
    class Linear(nnx.Module):
      def __init__(self, din: int, dout: int, *, rngs: nnx.Rngs):
        key = rngs.params()
        self.w = nnx.Param(jax.random.uniform(key, (din, dout)))
        self.b = nnx.Param(jnp.zeros((dout,)))

      def __call__(self, x):
        y = x @ self.w.value + self.b.value[None]
        self.y = nnx.Intermediate(y)
        return y

    model = Linear(12, 2, rngs=nnx.Rngs(0))

    y = model(jnp.ones((8, 12)))

    intermediates = nnx.pop(model, nnx.Intermediate)

    assert 'y' in intermediates

  def test_intermediates_example_functional(self):
    class Linear(nnx.Module):
      def __init__(self, din: int, dout: int, *, rngs: nnx.Rngs):
        key = rngs.params()
        self.w = nnx.Param(jax.random.uniform(key, (din, dout)))
        self.b = nnx.Param(jnp.zeros((dout,)))

      def __call__(self, x):
        y = x @ self.w.value + self.b.value[None]
        self.y = nnx.Intermediate(y)
        return y

    model = Linear(12, 2, rngs=nnx.Rngs(0))

    graphdef, state = nnx.split(model)

    y, (_, state) = graphdef.apply(state)(jnp.ones((8, 12)))

    intermediates, state = state.split(nnx.Intermediate, ...)

    assert 'y' in intermediates


if __name__ == '__main__':
  absltest.main()
